# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import base64
from unittest import mock

import msgpack
from autobahn.websocket.types import ConnectionDeny
from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.worker.protocols.manager.msgpack import BuildbotWebSocketServerProtocol
from buildbot.worker.protocols.manager.msgpack import ConnectioLostError
from buildbot.worker.protocols.manager.msgpack import RemoteWorkerError
from buildbot.worker.protocols.manager.msgpack import decode_http_authorization_header
from buildbot.worker.protocols.manager.msgpack import encode_http_authorization_header


class TestHttpAuthorization(unittest.TestCase):
    def test_encode(self):
        result = encode_http_authorization_header(b'name', b'pass')
        self.assertEqual(result, 'Basic bmFtZTpwYXNz')

        result = encode_http_authorization_header(b'name2', b'pass2')
        self.assertEqual(result, 'Basic bmFtZTI6cGFzczI=')

    def test_encode_username_contains_colon(self):
        with self.assertRaises(ValueError):
            encode_http_authorization_header(b'na:me', b'pass')

    def test_decode(self):
        result = decode_http_authorization_header(
            encode_http_authorization_header(b'name', b'pass')
        )
        self.assertEqual(result, ('name', 'pass'))

        # password can contain a colon
        result = decode_http_authorization_header(
            encode_http_authorization_header(b'name', b'pa:ss')
        )
        self.assertEqual(result, ('name', 'pa:ss'))

    def test_contains_no__basic(self):
        with self.assertRaises(ValueError):
            decode_http_authorization_header('Test bmFtZTpwYXNzOjI=')

        with self.assertRaises(ValueError):
            decode_http_authorization_header('TestTest bmFtZTpwYXNzOjI=')

    def test_contains_forbidden_character(self):
        with self.assertRaises(ValueError):
            decode_http_authorization_header('Basic test%test')

    def test_credentials_do_not_contain_colon(self):
        value = 'Basic ' + base64.b64encode(b'TestTestTest').decode()
        with self.assertRaises(ValueError):
            decode_http_authorization_header(value)


class TestException(Exception):
    pass


class TestBuildbotWebSocketServerProtocol(unittest.TestCase):
    def setUp(self):
        self.protocol = BuildbotWebSocketServerProtocol()
        self.protocol.sendMessage = mock.Mock()
        self.seq_number = 1

    @defer.inlineCallbacks
    def send_msg_check_response(self, protocol, msg, expected):
        msg = msg.copy()
        msg['seq_number'] = self.seq_number

        expected = expected.copy()
        expected['seq_number'] = self.seq_number
        self.seq_number += 1

        protocol.onMessage(msgpack.packb(msg), True)
        yield protocol._deferwaiter.wait()
        args_tuple, _ = protocol.sendMessage.call_args
        result = msgpack.unpackb(args_tuple[0], raw=False)
        self.assertEqual(result, expected)

    def send_msg_get_result(self, msg):
        msg = msg.copy()
        msg['seq_number'] = self.seq_number
        self.seq_number += 1

        self.protocol.onMessage(msgpack.packb(msg), True)

        args_tuple, _ = self.protocol.sendMessage.call_args
        return msgpack.unpackb(args_tuple[0], raw=False)['result']

    @defer.inlineCallbacks
    def connect_authenticated_worker(self):
        # worker has to be authenticated before opening the connection
        pfactory = mock.Mock()
        pfactory.connection = mock.Mock()

        self.setup_mock_users({'name': ('pass', pfactory)})

        request = mock.Mock()
        request.headers = {"authorization": 'Basic bmFtZTpwYXNz'}
        request.peer = ''

        yield self.protocol.onConnect(request)
        yield self.protocol.onOpen()

    def setup_mock_users(self, users):
        self.protocol.factory = mock.Mock()
        self.protocol.factory.buildbot_dispatcher = mock.Mock()
        self.protocol.factory.buildbot_dispatcher.users = users

    @parameterized.expand([
        ('update_op', {'seq_number': 1}),
        ('update_seq_number', {'op': 'update'}),
        ('complete_op', {'seq_number': 1}),
        ('complete_seq_number', {'op': 'complete'}),
        ('update_upload_file_write_op', {'seq_number': 1}),
        ('update_upload_file_write_seq_number', {'op': 'update_upload_file_write'}),
        ('update_upload_file_utime_op', {'seq_number': 1}),
        ('update_upload_file_utime_seq_number', {'op': 'update_upload_file_utime'}),
        ('update_upload_file_close_op', {'seq_number': 1}),
        ('update_upload_file_close_seq_number', {'op': 'update_upload_file_close'}),
        ('update_read_file_op', {'seq_number': 1}),
        ('update_read_file_seq_number', {'op': 'update_read_file'}),
        ('update_read_file_close_op', {'seq_number': 1}),
        ('update_read_file_close_seq_number', {'op': 'update_read_file_close'}),
        ('update_upload_directory_unpack_op', {'seq_number': 1}),
        ('update_upload_directory_unpack_seq_number', {'op': 'update_upload_directory_unpack'}),
        ('update_upload_directory_write_op', {'seq_number': 1}),
        ('update_upload_directory_write_seq_number', {'op': 'update_upload_directory_write'}),
    ])
    def test_msg_missing_arg(self, name, msg):
        with mock.patch('twisted.python.log.msg') as mock_log:
            self.protocol.onMessage(msgpack.packb(msg), True)
            mock_log.assert_any_call(f'Invalid message from worker: {msg}')

        # if msg does not have 'sep_number' or 'op', response sendMessage should not be called
        self.protocol.sendMessage.assert_not_called()

    @parameterized.expand([
        ('update', {'op': 'update', 'args': 'args'}),
        ('complete', {'op': 'complete', 'args': 'args'}),
        ('update_upload_file_write', {'op': 'update_upload_file_write', 'args': 'args'}),
        (
            'update_upload_file_utime',
            {'op': 'update_upload_file_utime', 'access_time': 1, 'modified_time': 2},
        ),
        ('update_upload_file_close', {'op': 'update_upload_file_close'}),
        ('update_read_file', {'op': 'update_read_file', 'length': 1}),
        ('update_read_file_close', {'op': 'update_read_file_close'}),
        ('update_upload_directory_unpack', {'op': 'update_upload_directory_unpack'}),
        ('upload_directory_write', {'op': 'update_upload_directory_write', 'args': 'args'}),
    ])
    @defer.inlineCallbacks
    def test_missing_command_id(self, command, msg):
        yield self.connect_authenticated_worker()
        expected = {
            'op': 'response',
            'result': '\'message did not contain obligatory "command_id" key\'',
            'is_exception': True,
        }
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @parameterized.expand([
        ('update', {'op': 'update', 'args': 'args', 'command_id': 2}, {1: 'remoteCommand'}),
        ('complete', {'op': 'complete', 'args': 'args', 'command_id': 2}, {1: 'remoteCommand'}),
    ])
    @defer.inlineCallbacks
    def test_unknown_command_id(self, command, msg, command_id_to_command_map):
        yield self.connect_authenticated_worker()
        self.protocol.command_id_to_command_map = command_id_to_command_map
        expected = {'op': 'response', 'result': '\'unknown "command_id"\'', 'is_exception': True}
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @parameterized.expand([
        (
            'update_upload_file_write',
            {'op': 'update_upload_file_write', 'args': 'args', 'command_id': 2},
        ),
        (
            'update_upload_directory_unpack',
            {'op': 'update_upload_directory_unpack', 'command_id': 2},
        ),
        ('update_upload_file_close', {'op': 'update_upload_file_close', 'command_id': 2}),
        (
            'update_upload_file_utime',
            {
                'op': 'update_upload_file_utime',
                'access_time': 1,
                'modified_time': 2,
                'command_id': 2,
            },
        ),
        (
            'update_upload_directory_write',
            {'op': 'update_upload_directory_write', 'command_id': 2, 'args': 'args'},
        ),
    ])
    @defer.inlineCallbacks
    def test_unknown_command_id_writers(self, command, msg):
        yield self.connect_authenticated_worker()
        self.protocol.command_id_to_writer_map = {1: 'writer'}
        expected = {'op': 'response', 'result': '\'unknown "command_id"\'', 'is_exception': True}
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @parameterized.expand([
        ('update', {'op': 'update', 'command_id': 2}),
        ('complete', {'op': 'complete', 'command_id': 2}),
        ('update_upload_file_write', {'op': 'update_upload_file_write', 'command_id': 2}),
        ('update_upload_directory_write', {'op': 'update_upload_directory_write', 'command_id': 1}),
    ])
    @defer.inlineCallbacks
    def test_missing_args(self, command, msg):
        yield self.connect_authenticated_worker()
        expected = {
            'op': 'response',
            'result': '\'message did not contain obligatory "args" key\'',
            'is_exception': True,
        }
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @parameterized.expand([
        ('update_read_file', {'op': 'update_read_file', 'length': 1, 'command_id': 2}),
        ('update_read_file_close', {'op': 'update_read_file_close', 'command_id': 2}),
    ])
    @defer.inlineCallbacks
    def test_unknown_command_id_readers(self, command, msg):
        yield self.connect_authenticated_worker()
        self.protocol.command_id_to_reader_map = {1: 'reader'}
        expected = {'op': 'response', 'result': '\'unknown "command_id"\'', 'is_exception': True}
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @defer.inlineCallbacks
    def test_missing_authorization_header(self):
        request = mock.Mock()
        request.headers = {"authorization": ''}
        request.peer = ''

        with self.assertRaises(ConnectionDeny):
            yield self.protocol.onConnect(request)

    @defer.inlineCallbacks
    def test_auth_password_does_not_match(self):
        pfactory = mock.Mock()
        pfactory.connection = mock.Mock()

        self.setup_mock_users({'username': ('password', pfactory)})

        request = mock.Mock()
        request.headers = {
            "authorization": encode_http_authorization_header(b'username', b'wrong_password')
        }
        request.peer = ''

        with self.assertRaises(ConnectionDeny):
            yield self.protocol.onConnect(request)

    @defer.inlineCallbacks
    def test_auth_username_unknown(self):
        pfactory = mock.Mock()
        pfactory.connection = mock.Mock()

        self.setup_mock_users({'username': ('pass', pfactory)})

        request = mock.Mock()

        request.headers = {
            "authorization": encode_http_authorization_header(b'wrong_username', b'pass')
        }
        request.peer = ''

        with self.assertRaises(ConnectionDeny):
            yield self.protocol.onConnect(request)

    @defer.inlineCallbacks
    def test_update_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_command_map = {command_id: command}

        msg = {'op': 'update', 'args': 'args', 'command_id': command_id}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_update_msgpack.assert_called_once_with(msg['args'])

    @defer.inlineCallbacks
    def test_complete_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_command_map = {command_id: command}
        self.protocol.command_id_to_reader_map = {}
        self.protocol.command_id_to_writer_map = {}

        msg = {'op': 'complete', 'args': 'args', 'command_id': command_id}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_complete.assert_called_once()

    @defer.inlineCallbacks
    def test_complete_check_dict_removal(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_command_map = {command_id: command, 2: 'test_command'}
        self.protocol.command_id_to_reader_map = {command_id: 'test_reader', 2: 'test_reader2'}
        self.protocol.command_id_to_writer_map = {command_id: 'test_writer', 2: 'test_writer2'}

        msg = {'op': 'complete', 'args': 'args', 'command_id': command_id}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_complete.assert_called_once()
        self.assertEqual(self.protocol.command_id_to_command_map, {2: 'test_command'})
        self.assertEqual(self.protocol.command_id_to_reader_map, {2: 'test_reader2'})
        self.assertEqual(self.protocol.command_id_to_writer_map, {2: 'test_writer2'})

    @defer.inlineCallbacks
    def test_update_upload_file_write_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_writer_map = {command_id: command}

        msg = {'op': 'update_upload_file_write', 'args': 'args', 'command_id': command_id}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_write.assert_called_once()

    @defer.inlineCallbacks
    def test_update_upload_file_utime_missing_access_time(self):
        yield self.connect_authenticated_worker()
        msg = {'op': 'update_upload_file_utime', 'modified_time': 2, 'command_id': 2}
        expected = {
            'op': 'response',
            'result': '\'message did not contain obligatory "access_time" key\'',
            'is_exception': True,
        }
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @defer.inlineCallbacks
    def test_update_upload_file_utime_missing_modified_time(self):
        yield self.connect_authenticated_worker()
        msg = {'op': 'update_upload_file_utime', 'access_time': 1, 'command_id': 2}
        expected = {
            'op': 'response',
            'result': '\'message did not contain obligatory "modified_time" key\'',
            'is_exception': True,
        }
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @defer.inlineCallbacks
    def test_update_upload_file_utime_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_writer_map = {command_id: command}

        msg = {
            'op': 'update_upload_file_utime',
            'access_time': 1,
            'modified_time': 2,
            'command_id': command_id,
        }
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_utime.assert_called_once_with('access_time', 'modified_time')

    @defer.inlineCallbacks
    def test_update_upload_file_close_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_writer_map = {command_id: command}

        msg = {'op': 'update_upload_file_close', 'command_id': command_id}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_close.assert_called_once()

    @defer.inlineCallbacks
    def test_update_read_file_missing_length(self):
        yield self.connect_authenticated_worker()
        msg = {'op': 'update_read_file', 'command_id': 1}
        expected = {
            'op': 'response',
            'result': '\'message did not contain obligatory "length" key\'',
            'is_exception': True,
        }
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @defer.inlineCallbacks
    def test_update_read_file_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_reader_map = {command_id: command}

        msg = {'op': 'update_read_file', 'length': 1, 'command_id': command_id}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_read.assert_called_once_with(msg['length'])

    @defer.inlineCallbacks
    def test_update_read_file_close_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_reader_map = {command_id: command}

        msg = {'op': 'update_read_file_close', 'command_id': command_id}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_close.assert_called_once()

    @defer.inlineCallbacks
    def test_update_upload_directory_unpack_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_writer_map = {command_id: command}

        msg = {'op': 'update_upload_directory_unpack', 'command_id': command_id}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_unpack.assert_called_once()

    @defer.inlineCallbacks
    def test_update_upload_directory_write_success(self):
        yield self.connect_authenticated_worker()
        command_id = 1

        command = mock.Mock()
        self.protocol.command_id_to_writer_map = {command_id: command}

        msg = {'op': 'update_upload_directory_write', 'command_id': command_id, 'args': 'args'}
        expected = {'op': 'response', 'result': None}
        yield self.send_msg_check_response(self.protocol, msg, expected)
        command.remote_write.assert_called_once_with(msg['args'])

    def test_onMessage_not_isBinary(self):
        # if isBinary is False, sendMessage should not be called
        msg = {}
        self.protocol.onMessage(msgpack.packb(msg), False)
        self.seq_number += 1
        self.protocol.sendMessage.assert_not_called()

    @defer.inlineCallbacks
    def test_onMessage_worker_not_authenticated(self):
        msg = {'op': 'update', 'command_id': 1, 'args': 'test'}
        expected = {
            'op': 'response',
            'result': 'Worker not authenticated.',
            'is_exception': True,
        }
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @defer.inlineCallbacks
    def test_onMessage_command_does_not_exist(self):
        yield self.connect_authenticated_worker()
        msg = {'op': 'test'}
        expected = {
            'op': 'response',
            'result': 'Command test does not exist.',
            'is_exception': True,
        }
        yield self.send_msg_check_response(self.protocol, msg, expected)

    @defer.inlineCallbacks
    def test_get_message_result_success(self):
        yield self.connect_authenticated_worker()
        msg = {'op': 'getWorkerInfo'}
        d = self.protocol.get_message_result(msg)
        seq_num = msg['seq_number']
        self.assertEqual(d.called, False)

        self.protocol.sendMessage.assert_called()

        # master got an answer from worker through onMessage
        msg = {'seq_number': seq_num, 'op': 'response', 'result': 'test_result'}
        self.protocol.onMessage(msgpack.packb(msg), isBinary=True)
        self.assertEqual(d.called, True)
        res = yield d
        self.assertEqual(res, 'test_result')

    @defer.inlineCallbacks
    def test_get_message_result_failed(self):
        yield self.connect_authenticated_worker()
        msg = {'op': 'getWorkerInfo'}
        d = self.protocol.get_message_result(msg)
        seq_num = msg['seq_number']
        self.assertEqual(d.called, False)

        # Master got an answer from worker through onMessage.
        # This time the message indicates failure
        msg_response = {
            'seq_number': seq_num,
            'op': 'response',
            'is_exception': True,
            'result': 'error_result',
        }
        self.protocol.onMessage(msgpack.packb(msg_response), isBinary=True)
        self.assertEqual(d.called, True)
        with self.assertRaises(RemoteWorkerError):
            yield d

    @defer.inlineCallbacks
    def test_get_message_result_no_worker_connection(self):
        # master can not send any messages if connection is not established
        with self.assertRaises(ConnectioLostError):
            yield self.protocol.get_message_result({'op': 'getWorkerInfo'})

    @defer.inlineCallbacks
    def test_onClose_connection_lost_error(self):
        yield self.connect_authenticated_worker()
        # master sends messages for worker and waits for their response
        msg = {'op': 'getWorkerInfo'}
        d1 = self.protocol.get_message_result(msg)
        self.assertEqual(d1.called, False)

        msg = {'op': 'print', 'message': 'test'}
        d2 = self.protocol.get_message_result(msg)
        self.assertEqual(d2.called, False)

        # Worker disconnected, master will never get the response message.
        # Stop waiting and raise Exception
        self.protocol.onClose(True, None, 'worker is gone')
        self.assertEqual(d1.called, True)
        with self.assertRaises(ConnectioLostError):
            yield d1

        self.assertEqual(d2.called, True)
        with self.assertRaises(ConnectioLostError):
            yield d2

        self.protocol.connection.detached.assert_called()
        # contents of dict_def are deleted to stop waiting for the responses of all commands
        self.assertEqual(len(self.protocol.seq_num_to_waiters_map), 0)



#! /usr/bin/python


import os.path
import random
import subprocess
import sys

from twisted.cred import credentials
from twisted.internet import reactor
from twisted.python import log
from twisted.spread import pb

"""
This is an example of how to use the remote ChangeMaster interface, which is
a port that allows a remote program to inject Changes into the buildmaster.

The buildmaster can either pull changes in from external sources (see
buildbot.changes.changes.ChangeMaster.addSource for an example), or those
changes can be pushed in from outside. This script shows how to do the
pushing.

Changes are just dictionaries with three keys:

 'who': a simple string with a username. Responsibility for this change will
 be assigned to the named user (if something goes wrong with the build, they
 will be blamed for it).

 'files': a list of strings, each with a filename relative to the top of the
 source tree.

 'comments': a (multiline) string with checkin comments.

Each call to .addChange injects a single Change object: each Change
represents multiple files, all changed by the same person, and all with the
same checkin comments.

The port that this script connects to is the same 'workerPort' that the
workers and other debug tools use. The ChangeMaster service will only be
available on that port if 'change' is in the list of services passed to
buildbot.master.makeApp (this service is turned ON by default).
"""


def done(*args):
    reactor.stop()


users = ('zaphod', 'arthur', 'trillian', 'marvin', 'sbfast')
dirs = ('src', 'doc', 'tests')
sources = ('foo.c', 'bar.c', 'baz.c', 'Makefile')
docs = ('Makefile', 'index.html', 'manual.texinfo')


def makeFilename():
    d = random.choice(dirs)
    if d in ('src', 'tests'):
        f = random.choice(sources)
    else:
        f = random.choice(docs)
    return os.path.join(d, f)


def send_change(remote):
    who = random.choice(users)
    if len(sys.argv) > 1:
        files = sys.argv[1:]
    else:
        files = [makeFilename()]
    comments = subprocess.check_output(["fortune"])
    comments = comments.decode(sys.stdout.encoding)
    change = {'who': who, 'files': files, 'comments': comments}
    d = remote.callRemote('addChange', change)
    d.addCallback(done)
    print("{}: {}".format(who, " ".join(files)))


f = pb.PBClientFactory()
d = f.login(credentials.UsernamePassword("change", "changepw"))
reactor.connectTCP("10.0.24.125", 9989, f)
err = lambda f: (log.err(), reactor.stop())
d.addCallback(send_change).addErrback(err)

reactor.run()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass

import sqlalchemy as sa
from twisted.internet import defer

from buildbot.db import NULL
from buildbot.db import base
from buildbot.util.sautils import hash_columns
from buildbot.warnings import warn_deprecated


class ChangeSourceAlreadyClaimedError(Exception):
    pass


@dataclass
class ChangeSourceModel:
    id: int
    name: str

    masterid: int | None = None

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'ChangeSourcesConnectorComponent '
                'getChangeSource, and getChangeSources '
                'no longer return ChangeSource as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


class ChangeSourcesConnectorComponent(base.DBConnectorComponent):
    def findChangeSourceId(self, name):
        tbl = self.db.model.changesources
        name_hash = hash_columns(name)
        return self.findSomethingId(
            tbl=tbl,
            whereclause=(tbl.c.name_hash == name_hash),
            insert_values={"name": name, "name_hash": name_hash},
        )

    # returns a Deferred that returns None
    def setChangeSourceMaster(self, changesourceid, masterid):
        def thd(conn):
            cs_mst_tbl = self.db.model.changesource_masters

            # handle the masterid=None case to get it out of the way
            if masterid is None:
                q = cs_mst_tbl.delete().where(cs_mst_tbl.c.changesourceid == changesourceid)
                conn.execute(q)
                conn.commit()
                return

            # try a blind insert..
            try:
                q = cs_mst_tbl.insert()
                conn.execute(q, {"changesourceid": changesourceid, "masterid": masterid})
                conn.commit()
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError) as e:
                conn.rollback()
                # someone already owns this changesource.
                raise ChangeSourceAlreadyClaimedError from e

        return self.db.pool.do(thd)

    def get_change_source_master(self, changesourceid):
        def thd(conn):
            q = sa.select(self.db.model.changesource_masters.c.masterid).where(
                self.db.model.changesource_masters.c.changesourceid == changesourceid
            )
            r = conn.execute(q)
            row = r.fetchone()
            conn.close()
            if row:
                return row.masterid
            return None

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def getChangeSource(self, changesourceid):
        cs = yield self.getChangeSources(_changesourceid=changesourceid)
        if cs:
            return cs[0]
        return None

    # returns a Deferred that returns a value
    def getChangeSources(self, active=None, masterid=None, _changesourceid=None):
        def thd(conn):
            cs_tbl = self.db.model.changesources
            cs_mst_tbl = self.db.model.changesource_masters

            # handle the trivial case of masterid=xx and active=False
            if masterid is not None and active is not None and not active:
                return []

            join = cs_tbl.outerjoin(cs_mst_tbl, (cs_tbl.c.id == cs_mst_tbl.c.changesourceid))

            # if we're given a _changesourceid, select only that row
            wc = None
            if _changesourceid:
                wc = cs_tbl.c.id == _changesourceid
            else:
                # otherwise, filter with active, if necessary
                if masterid is not None:
                    wc = cs_mst_tbl.c.masterid == masterid
                elif active:
                    wc = cs_mst_tbl.c.masterid != NULL
                elif active is not None:
                    wc = cs_mst_tbl.c.masterid == NULL

            q = sa.select(
                cs_tbl.c.id,
                cs_tbl.c.name,
                cs_mst_tbl.c.masterid,
            ).select_from(join)
            if wc is not None:
                q = q.where(wc)

            return [self._model_from_row(row) for row in conn.execute(q).fetchall()]

        return self.db.pool.do(thd)

    def _model_from_row(self, row):
        return ChangeSourceModel(id=row.id, name=row.name, masterid=row.masterid)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.config import ConfigErrors
from buildbot.process.properties import Property
from buildbot.process.results import SUCCESS
from buildbot.steps.cmake import CMake
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin


class TestCMake(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setup_test_build_step()

    def expect_and_run_command(self, *params):
        command = [CMake.DEFAULT_CMAKE, *list(params)]

        self.expect_commands(ExpectShell(command=command, workdir='wkdir').exit(0))
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_definitions_type(self):
        with self.assertRaises(ConfigErrors):
            CMake(definitions='hello')

    def test_options_type(self):
        with self.assertRaises(ConfigErrors):
            CMake(options='hello')

    def test_plain(self):
        self.setup_step(CMake())
        self.expect_commands(ExpectShell(command=[CMake.DEFAULT_CMAKE], workdir='wkdir').exit(0))
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_cmake(self):
        cmake_bin = 'something/else/cmake'

        self.setup_step(CMake(cmake=cmake_bin))
        self.expect_commands(ExpectShell(command=[cmake_bin], workdir='wkdir').exit(0))
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_cmake_interpolation(self):
        prop = 'CMAKE'
        value = 'Real_CMAKE'

        self.setup_step(CMake(cmake=Property(prop)))
        self.build.setProperty(prop, value, source='test')

        self.expect_commands(ExpectShell(command=[value], workdir='wkdir').exit(0))
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_definitions(self):
        definition = {'a': 'b'}
        self.setup_step(CMake(definitions=definition))
        self.expect_and_run_command('-Da=b')

    def test_environment(self):
        command = [CMake.DEFAULT_CMAKE]
        environment = {'a': 'b'}
        self.setup_step(CMake(env=environment))
        self.expect_commands(ExpectShell(command=command, workdir='wkdir', env={'a': 'b'}).exit(0))
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_definitions_interpolation(self):
        definitions = {'a': Property('b')}

        self.setup_step(CMake(definitions=definitions))
        self.build.setProperty('b', 'real_b', source='test')
        self.expect_and_run_command('-Da=real_b')

    def test_definitions_renderable(self):
        definitions = Property('b')
        self.setup_step(CMake(definitions=definitions))
        self.build.setProperty('b', {'a': 'real_b'}, source='test')
        self.expect_and_run_command('-Da=real_b')

    def test_generator(self):
        generator = 'Ninja'

        self.setup_step(CMake(generator=generator))
        self.expect_and_run_command('-G', generator)

    def test_generator_interpolation(self):
        value = 'Our_GENERATOR'

        self.setup_step(CMake(generator=Property('GENERATOR')))
        self.build.setProperty('GENERATOR', value, source='test')

        self.expect_and_run_command('-G', value)

    def test_options(self):
        options = ('A', 'B')

        self.setup_step(CMake(options=options))
        self.expect_and_run_command(*options)

    def test_options_interpolation(self):
        prop = 'option'
        value = 'value'

        self.setup_step(CMake(options=(Property(prop),)))
        self.build.setProperty(prop, value, source='test')
        self.expect_and_run_command(value)

    def test_path(self):
        path = 'some/path'

        self.setup_step(CMake(path=path))
        self.expect_and_run_command(path)

    def test_path_interpolation(self):
        prop = 'path'
        value = 'some/path'

        self.setup_step(CMake(path=Property(prop)))
        self.build.setProperty(prop, value, source='test')
        self.expect_and_run_command(value)

    def test_options_path(self):
        self.setup_step(CMake(path='some/path', options=('A', 'B')))
        self.expect_and_run_command('A', 'B', 'some/path')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import pprint
from typing import TYPE_CHECKING

from buildbot_worker.base import ProtocolCommandBase

if TYPE_CHECKING:
    from typing import Any
    from typing import Iterable

    from twisted.internet.defer import Deferred

    from buildbot_worker.test.fake.remote import FakeRemote


class FakeProtocolCommand(ProtocolCommandBase):
    debug = False

    def __init__(self, basedir: str) -> None:
        self.unicode_encoding = 'utf-8'
        self.updates: list[tuple[str, Any] | str] = []
        self.worker_basedir = basedir
        self.basedir = basedir

    def show(self) -> str:
        return pprint.pformat(self.updates)

    def send_update(self, status: Iterable[tuple[str, Any]]) -> None:
        if self.debug:
            print("FakeWorkerForBuilder.sendUpdate", status)
        for st in status:
            self.updates.append(st)

    def protocol_update_upload_file_close(self, writer: FakeRemote) -> Deferred[None]:
        return writer.callRemote("close")

    def protocol_update_upload_file_utime(
        self,
        writer: FakeRemote,
        access_time: float,
        modified_time: float,
    ) -> Deferred[None]:
        return writer.callRemote("utime", (access_time, modified_time))

    def protocol_update_upload_file_write(
        self,
        writer: FakeRemote,
        data: str | bytes,
    ) -> Deferred[None]:
        return writer.callRemote('write', data)

    def protocol_update_upload_directory(self, writer: FakeRemote) -> Deferred[None]:
        return writer.callRemote("unpack")

    def protocol_update_upload_directory_write(
        self,
        writer: FakeRemote,
        data: str | bytes,
    ) -> Deferred[None]:
        return writer.callRemote('write', data)

    def protocol_update_read_file_close(self, reader: FakeRemote) -> Deferred[None]:
        return reader.callRemote('close')

    def protocol_update_read_file(self, reader: FakeRemote, length: int) -> Deferred[None]:
        return reader.callRemote('read', length)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import schedulers
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin


class Tests(TestReactorMixin, unittest.TestCase):
    # test data

    ss92 = fakedb.SourceStamp(id=92)
    change3 = fakedb.Change(changeid=3)
    change4 = fakedb.Change(changeid=4)
    change5 = fakedb.Change(changeid=5)
    change6 = fakedb.Change(changeid=6, branch='sql')

    scheduler24 = fakedb.Scheduler(id=24, name='schname')
    master13 = fakedb.Master(id=13, name='m1', active=1)
    scheduler24master = fakedb.SchedulerMaster(schedulerid=24, masterid=13)

    scheduler25 = fakedb.Scheduler(id=25, name='schname2')
    master14 = fakedb.Master(id=14, name='m2', active=0)
    scheduler25master = fakedb.SchedulerMaster(schedulerid=25, masterid=14)

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_enable(self):
        yield self.db.insert_test_data([self.scheduler24, self.master13, self.scheduler24master])
        sch = yield self.db.schedulers.getScheduler(24)
        self.assertIsInstance(sch, schedulers.SchedulerModel)
        self.assertEqual(
            sch, schedulers.SchedulerModel(id=24, name='schname', enabled=True, masterid=13)
        )

        yield self.db.schedulers.enable(24, False)
        sch = yield self.db.schedulers.getScheduler(24)
        self.assertIsInstance(sch, schedulers.SchedulerModel)
        self.assertEqual(
            sch, schedulers.SchedulerModel(id=24, name='schname', enabled=False, masterid=13)
        )

        yield self.db.schedulers.enable(24, True)
        sch = yield self.db.schedulers.getScheduler(24)
        self.assertIsInstance(sch, schedulers.SchedulerModel)
        self.assertEqual(
            sch, schedulers.SchedulerModel(id=24, name='schname', enabled=True, masterid=13)
        )

    @defer.inlineCallbacks
    def test_classifyChanges(self):
        yield self.db.insert_test_data([self.ss92, self.change3, self.change4, self.scheduler24])
        yield self.db.schedulers.classifyChanges(24, {3: False, 4: True})
        res = yield self.db.schedulers.getChangeClassifications(24)
        self.assertEqual(res, {3: False, 4: True})

    @defer.inlineCallbacks
    def test_classifyChanges_again(self):
        # test reclassifying changes, which may happen during some timing
        # conditions.  It's important that this test uses multiple changes,
        # only one of which already exists
        yield self.db.insert_test_data([
            self.ss92,
            self.change3,
            self.change4,
            self.change5,
            self.change6,
            self.scheduler24,
            fakedb.SchedulerChange(schedulerid=24, changeid=5, important=0),
        ])
        yield self.db.schedulers.classifyChanges(24, {3: True, 4: False, 5: True, 6: False})
        res = yield self.db.schedulers.getChangeClassifications(24)
        self.assertEqual(res, {3: True, 4: False, 5: True, 6: False})

    @defer.inlineCallbacks
    def test_flushChangeClassifications(self):
        yield self.db.insert_test_data([
            self.ss92,
            self.change3,
            self.change4,
            self.change5,
            self.scheduler24,
        ])
        yield self.addClassifications(24, (3, 1), (4, 0), (5, 1))
        res = yield self.db.schedulers.getChangeClassifications(24)
        self.assertEqual(res, {3: True, 4: False, 5: True})
        yield self.db.schedulers.flushChangeClassifications(24)
        res = yield self.db.schedulers.getChangeClassifications(24)
        self.assertEqual(res, {})

    @defer.inlineCallbacks
    def test_flushChangeClassifications_less_than(self):
        yield self.db.insert_test_data([
            self.ss92,
            self.change3,
            self.change4,
            self.change5,
            self.scheduler24,
        ])
        yield self.addClassifications(24, (3, 1), (4, 0), (5, 1))
        yield self.db.schedulers.flushChangeClassifications(24, less_than=5)
        res = yield self.db.schedulers.getChangeClassifications(24)
        self.assertEqual(res, {5: True})

    @defer.inlineCallbacks
    def test_getChangeClassifications(self):
        yield self.db.insert_test_data([
            self.ss92,
            self.change3,
            self.change4,
            self.change5,
            self.change6,
            self.scheduler24,
        ])
        yield self.addClassifications(24, (3, 1), (4, 0), (5, 1), (6, 1))
        res = yield self.db.schedulers.getChangeClassifications(24)
        self.assertEqual(res, {3: True, 4: False, 5: True, 6: True})

    @defer.inlineCallbacks
    def test_getChangeClassifications_branch(self):
        yield self.db.insert_test_data([
            self.ss92,
            self.change3,
            self.change4,
            self.change5,
            self.change6,
            self.scheduler24,
        ])
        yield self.addClassifications(24, (3, 1), (4, 0), (5, 1), (6, 1))
        res = yield self.db.schedulers.getChangeClassifications(24, branch='sql')
        self.assertEqual(res, {6: True})

    @defer.inlineCallbacks
    def test_findSchedulerId_new(self):
        id = yield self.db.schedulers.findSchedulerId('schname')
        sch = yield self.db.schedulers.getScheduler(id)
        self.assertEqual(sch.name, 'schname')

    @defer.inlineCallbacks
    def test_findSchedulerId_existing(self):
        id = yield self.db.schedulers.findSchedulerId('schname')
        id2 = yield self.db.schedulers.findSchedulerId('schname')
        self.assertEqual(id, id2)

    @defer.inlineCallbacks
    def test_setSchedulerMaster_fresh(self):
        yield self.db.insert_test_data([self.scheduler24, self.master13])
        yield self.db.schedulers.setSchedulerMaster(24, 13)
        sch = yield self.db.schedulers.getScheduler(24)
        self.assertEqual(sch.masterid, 13)

    @defer.inlineCallbacks
    def test_setSchedulerMaster_inactive_but_linked(self):
        yield self.db.insert_test_data([
            self.master13,
            self.scheduler25,
            self.master14,
            self.scheduler25master,
        ])
        with self.assertRaises(schedulers.SchedulerAlreadyClaimedError):
            yield self.db.schedulers.setSchedulerMaster(25, 13)

    @defer.inlineCallbacks
    def test_setSchedulerMaster_inactive_but_linked_to_this_master(self):
        yield self.db.insert_test_data([
            self.scheduler25,
            self.master14,
            self.scheduler25master,
        ])
        yield self.db.schedulers.setSchedulerMaster(25, 14)

    @defer.inlineCallbacks
    def test_setSchedulerMaster_active(self):
        yield self.db.insert_test_data([
            self.scheduler24,
            self.master13,
            self.scheduler24master,
        ])
        with self.assertRaises(schedulers.SchedulerAlreadyClaimedError):
            yield self.db.schedulers.setSchedulerMaster(24, 14)

    @defer.inlineCallbacks
    def test_setSchedulerMaster_None(self):
        yield self.db.insert_test_data([
            self.scheduler25,
            self.master14,
            self.scheduler25master,
        ])
        yield self.db.schedulers.setSchedulerMaster(25, None)
        sch = yield self.db.schedulers.getScheduler(25)
        self.assertEqual(sch.masterid, None)

    @defer.inlineCallbacks
    def test_setSchedulerMaster_None_unowned(self):
        yield self.db.insert_test_data([self.scheduler25])
        yield self.db.schedulers.setSchedulerMaster(25, None)
        sch = yield self.db.schedulers.getScheduler(25)
        self.assertEqual(sch.masterid, None)

    @defer.inlineCallbacks
    def test_getScheduler(self):
        yield self.db.insert_test_data([self.scheduler24])
        sch = yield self.db.schedulers.getScheduler(24)
        self.assertIsInstance(sch, schedulers.SchedulerModel)
        self.assertEqual(
            sch, schedulers.SchedulerModel(id=24, name='schname', enabled=True, masterid=None)
        )

    @defer.inlineCallbacks
    def test_getScheduler_missing(self):
        sch = yield self.db.schedulers.getScheduler(24)
        self.assertEqual(sch, None)

    @defer.inlineCallbacks
    def test_getScheduler_active(self):
        yield self.db.insert_test_data([self.scheduler24, self.master13, self.scheduler24master])
        sch = yield self.db.schedulers.getScheduler(24)
        self.assertIsInstance(sch, schedulers.SchedulerModel)
        self.assertEqual(
            sch, schedulers.SchedulerModel(id=24, name='schname', enabled=True, masterid=13)
        )

    @defer.inlineCallbacks
    def test_getScheduler_inactive_but_linked(self):
        yield self.db.insert_test_data([self.scheduler25, self.master14, self.scheduler25master])
        sch = yield self.db.schedulers.getScheduler(25)
        self.assertIsInstance(sch, schedulers.SchedulerModel)
        self.assertEqual(
            sch, schedulers.SchedulerModel(id=25, name='schname2', enabled=True, masterid=14)
        )  # row exists, but marked inactive

    @defer.inlineCallbacks
    def test_getSchedulers(self):
        yield self.db.insert_test_data([
            self.scheduler24,
            self.master13,
            self.scheduler24master,
            self.scheduler25,
        ])

        def schKey(sch):
            return sch.id

        schlist = yield self.db.schedulers.getSchedulers()

        for sch in schlist:
            self.assertIsInstance(sch, schedulers.SchedulerModel)

        self.assertEqual(
            sorted(schlist, key=schKey),
            sorted(
                [
                    schedulers.SchedulerModel(id=24, name='schname', enabled=True, masterid=13),
                    schedulers.SchedulerModel(id=25, name='schname2', enabled=True, masterid=None),
                ],
                key=schKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getSchedulers_masterid(self):
        yield self.db.insert_test_data([
            self.scheduler24,
            self.master13,
            self.scheduler24master,
            self.scheduler25,
        ])
        schlist = yield self.db.schedulers.getSchedulers(masterid=13)

        for sch in schlist:
            self.assertIsInstance(sch, schedulers.SchedulerModel)

        self.assertEqual(
            sorted(schlist),
            sorted([
                schedulers.SchedulerModel(id=24, name='schname', enabled=True, masterid=13),
            ]),
        )

    @defer.inlineCallbacks
    def test_getSchedulers_active(self):
        yield self.db.insert_test_data([
            self.scheduler24,
            self.master13,
            self.scheduler24master,
            self.scheduler25,
        ])
        schlist = yield self.db.schedulers.getSchedulers(active=True)

        for sch in schlist:
            self.assertIsInstance(sch, schedulers.SchedulerModel)

        self.assertEqual(
            sorted(schlist),
            sorted([
                schedulers.SchedulerModel(id=24, name='schname', enabled=True, masterid=13),
            ]),
        )

    @defer.inlineCallbacks
    def test_getSchedulers_active_masterid(self):
        yield self.db.insert_test_data([
            self.scheduler24,
            self.master13,
            self.scheduler24master,
            self.scheduler25,
        ])
        schlist = yield self.db.schedulers.getSchedulers(active=True, masterid=13)

        for sch in schlist:
            self.assertIsInstance(sch, schedulers.SchedulerModel)

        self.assertEqual(
            sorted(schlist),
            sorted([
                schedulers.SchedulerModel(id=24, name='schname', enabled=True, masterid=13),
            ]),
        )

        schlist = yield self.db.schedulers.getSchedulers(active=True, masterid=14)

        for sch in schlist:
            self.assertIsInstance(sch, schedulers.SchedulerModel)

        self.assertEqual(sorted(schlist), [])

    @defer.inlineCallbacks
    def test_getSchedulers_inactive(self):
        yield self.db.insert_test_data([
            self.scheduler24,
            self.master13,
            self.scheduler24master,
            self.scheduler25,
        ])
        schlist = yield self.db.schedulers.getSchedulers(active=False)

        for sch in schlist:
            self.assertIsInstance(sch, schedulers.SchedulerModel)

        self.assertEqual(
            sorted(schlist),
            sorted([
                schedulers.SchedulerModel(id=25, name='schname2', enabled=True, masterid=None),
            ]),
        )

    @defer.inlineCallbacks
    def test_getSchedulers_inactive_masterid(self):
        yield self.db.insert_test_data([
            self.scheduler24,
            self.master13,
            self.scheduler24master,
            self.scheduler25,
        ])
        schlist = yield self.db.schedulers.getSchedulers(active=False, masterid=13)

        for sch in schlist:
            self.assertIsInstance(sch, schedulers.SchedulerModel)

        self.assertEqual(sorted(schlist), [])

        schlist = yield self.db.schedulers.getSchedulers(active=False, masterid=14)

        for sch in schlist:
            self.assertIsInstance(sch, schedulers.SchedulerModel)

        self.assertEqual(sorted(schlist), [])  # always returns [] by spec!

    @defer.inlineCallbacks
    def addClassifications(self, schedulerid, *classifications):
        def thd(conn):
            q = self.db.model.scheduler_changes.insert()
            conn.execute(
                q,
                [
                    {"changeid": c[0], "schedulerid": schedulerid, "important": c[1]}
                    for c in classifications
                ],
            )

        yield self.db.pool.do_with_transaction(thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.db import codebases
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util.twisted import async_to_deferred


class Tests(TestReactorMixin, unittest.TestCase):
    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        self.setup_test_reactor()
        self.master = await fakemaster.make_master(self, wantDb=True)
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Project(id=8, name='fake_project8'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1', slug='slug_codebase1'),
            fakedb.Codebase(id=14, projectid=7, name='codebase2'),
            fakedb.Codebase(id=15, projectid=7, name='codebase3'),
        ])

    @async_to_deferred
    async def test_find_codebase_id_exists(self) -> None:
        id = await self.master.db.codebases.find_codebase_id(projectid=7, name='codebase1')
        self.assertEqual(id, 13)

    @async_to_deferred
    async def test_find_codebase_id_new_no_auto_create(self) -> None:
        id = await self.master.db.codebases.find_codebase_id(
            projectid=7, name='codebase_not_exist', auto_create=False
        )
        self.assertIsNone(id)

    @async_to_deferred
    async def test_find_codebase_id_new(self) -> None:
        id = await self.master.db.codebases.find_codebase_id(projectid=7, name='codebase_not_exist')
        codebase = await self.master.db.codebases.get_codebase(id)
        self.assertEqual(
            codebase,
            codebases.CodebaseModel(
                id=id, projectid=7, name='codebase_not_exist', slug='codebase_not_exist'
            ),
        )

    @async_to_deferred
    async def test_get_codebase_exists(self) -> None:
        dbdict = await self.master.db.codebases.get_codebase(13)
        self.assertEqual(
            dbdict,
            codebases.CodebaseModel(id=13, projectid=7, name='codebase1', slug='slug_codebase1'),
        )

    @async_to_deferred
    async def test_get_codebase_does_not_exist(self) -> None:
        dbdict = await self.master.db.codebases.get_codebase(100)
        self.assertIsNone(dbdict)

    @async_to_deferred
    async def test_get_codebases(self) -> None:
        values = await self.master.db.codebases.get_codebases(projectid=7)
        values = sorted(values, key=lambda c: c.id)
        self.assertEqual(
            values,
            [
                codebases.CodebaseModel(
                    id=13, projectid=7, name='codebase1', slug='slug_codebase1'
                ),
                codebases.CodebaseModel(id=14, projectid=7, name='codebase2', slug='codebase2'),
                codebases.CodebaseModel(id=15, projectid=7, name='codebase3', slug='codebase3'),
            ],
        )

    @async_to_deferred
    async def test_update_codebase_info(self) -> None:
        await self.master.db.codebases.update_codebase_info(
            codebaseid=13, projectid=8, slug='new_slug'
        )
        dbdict = await self.master.db.codebases.get_codebase(13)
        self.assertEqual(
            dbdict,
            codebases.CodebaseModel(
                id=13,
                projectid=8,
                name="codebase1",
                slug="new_slug",
            ),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.trial import unittest

from buildbot.config.builder import BuilderConfig
from buildbot.process import factory
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.warnings import assertNotProducesWarnings
from buildbot.warnings import DeprecatedApiWarning


class BuilderConfigTests(ConfigErrorsMixin, unittest.TestCase):
    factory = factory.BuildFactory()

    # utils

    def assertAttributes(self, cfg, **expected):
        got = {attr: getattr(cfg, attr) for attr, exp in expected.items()}
        self.assertEqual(got, expected)

    # tests

    def test_no_name(self):
        with self.assertRaisesConfigError("builder's name is required"):
            BuilderConfig(factory=self.factory, workernames=['a'])

    def test_reserved_name(self):
        with self.assertRaisesConfigError("builder names must not start with an underscore: '_a'"):
            BuilderConfig(name='_a', factory=self.factory, workernames=['a'])

    def test_utf8_name(self):
        with self.assertRaisesConfigError("builder names must be unicode or ASCII"):
            BuilderConfig(name="\N{SNOWMAN}".encode(), factory=self.factory, workernames=['a'])

    def test_no_factory(self):
        with self.assertRaisesConfigError("builder 'a': has no factory"):
            BuilderConfig(name='a', workernames=['a'])

    def test_wrong_type_factory(self):
        with self.assertRaisesConfigError("builder 'a': factory is not"):
            BuilderConfig(factory=[], name='a', workernames=['a'])

    def test_no_workernames(self):
        with self.assertRaisesConfigError("builder 'a': at least one workername is required"):
            BuilderConfig(name='a', factory=self.factory)

    def test_bogus_workernames(self):
        with self.assertRaisesConfigError("workernames must be a list or a string"):
            BuilderConfig(name='a', workernames={1: 2}, factory=self.factory)

    def test_bogus_workername(self):
        with self.assertRaisesConfigError("workername must be a string"):
            BuilderConfig(name='a', workername=1, factory=self.factory)

    def test_tags_must_be_list(self):
        with self.assertRaisesConfigError("tags must be a list"):
            BuilderConfig(tags='abc', name='a', workernames=['a'], factory=self.factory)

    def test_tags_must_be_list_of_str(self):
        with self.assertRaisesConfigError("tags list contains something that is not a string"):
            BuilderConfig(tags=['abc', 13], name='a', workernames=['a'], factory=self.factory)

    def test_tags_no_tag_dupes(self):
        with self.assertRaisesConfigError("builder 'a': tags list contains duplicate tags: abc"):
            BuilderConfig(
                tags=['abc', 'bca', 'abc'], name='a', workernames=['a'], factory=self.factory
            )

    def test_inv_nextWorker(self):
        with self.assertRaisesConfigError("nextWorker must be a callable"):
            BuilderConfig(nextWorker="foo", name="a", workernames=['a'], factory=self.factory)

    def test_inv_nextBuild(self):
        with self.assertRaisesConfigError("nextBuild must be a callable"):
            BuilderConfig(nextBuild="foo", name="a", workernames=['a'], factory=self.factory)

    def test_inv_canStartBuild(self):
        with self.assertRaisesConfigError("canStartBuild must be a callable"):
            BuilderConfig(canStartBuild="foo", name="a", workernames=['a'], factory=self.factory)

    def test_inv_env(self):
        with self.assertRaisesConfigError("builder's env must be a dictionary"):
            BuilderConfig(env="foo", name="a", workernames=['a'], factory=self.factory)

    def test_defaults(self):
        cfg = BuilderConfig(name='a b c', workername='a', factory=self.factory)
        self.assertIdentical(cfg.factory, self.factory)
        self.assertAttributes(
            cfg,
            name='a b c',
            workernames=['a'],
            builddir='a_b_c',
            workerbuilddir='a_b_c',
            tags=[],
            nextWorker=None,
            locks=[],
            env={},
            properties={},
            collapseRequests=None,
            description=None,
        )

    def test_unicode_name(self):
        cfg = BuilderConfig(name='a \N{SNOWMAN} c', workername='a', factory=self.factory)
        self.assertIdentical(cfg.factory, self.factory)
        self.assertAttributes(cfg, name='a \N{SNOWMAN} c')

    def test_args(self):
        cfg = BuilderConfig(
            name='b',
            workername='s1',
            workernames='s2',
            builddir='bd',
            workerbuilddir='wbd',
            factory=self.factory,
            tags=['c'],
            nextWorker=lambda: 'ns',
            nextBuild=lambda: 'nb',
            locks=['l'],
            env={"x": 10},
            properties={"y": 20},
            collapseRequests='cr',
            description='buzz',
        )
        self.assertIdentical(cfg.factory, self.factory)
        self.assertAttributes(
            cfg,
            name='b',
            workernames=['s2', 's1'],
            builddir='bd',
            workerbuilddir='wbd',
            tags=['c'],
            locks=['l'],
            env={'x': 10},
            properties={'y': 20},
            collapseRequests='cr',
            description='buzz',
        )

    def test_too_long_property(self):
        with self.assertRaisesConfigError("exceeds maximum length of"):
            BuilderConfig(
                name="a", workernames=['a'], factory=self.factory, properties={'a' * 257: 'value'}
            )

    def test_too_long_default_property(self):
        with self.assertRaisesConfigError("exceeds maximum length of"):
            BuilderConfig(
                name="a",
                workernames=['a'],
                factory=self.factory,
                defaultProperties={'a' * 257: 'value'},
            )

    def test_description_wrong_format(self):
        with self.assertRaisesConfigError("builder description format must be None"):
            BuilderConfig(
                name="a", workernames=['a'], factory=self.factory, description_format="unknown"
            )

    def test_getConfigDict(self):
        ns = lambda: 'ns'
        nb = lambda: 'nb'
        cfg = BuilderConfig(
            name='b',
            workername='s1',
            workernames='s2',
            builddir='bd',
            workerbuilddir='wbd',
            factory=self.factory,
            tags=['c'],
            nextWorker=ns,
            nextBuild=nb,
            locks=['l'],
            env={"x": 10},
            properties={"y": 20},
            collapseRequests='cr',
            description='buzz',
        )
        self.assertEqual(
            cfg.getConfigDict(),
            {
                'builddir': 'bd',
                'tags': ['c'],
                'description': 'buzz',
                'env': {'x': 10},
                'factory': self.factory,
                'locks': ['l'],
                'collapseRequests': 'cr',
                'name': 'b',
                'nextBuild': nb,
                'nextWorker': ns,
                'properties': {'y': 20},
                'workerbuilddir': 'wbd',
                'workernames': ['s2', 's1'],
            },
        )

    def test_getConfigDict_collapseRequests(self):
        for cr in (False, lambda a, b, c: False):
            cfg = BuilderConfig(
                name='b', collapseRequests=cr, factory=self.factory, workername='s1'
            )
            self.assertEqual(
                cfg.getConfigDict(),
                {
                    'builddir': 'b',
                    'collapseRequests': cr,
                    'name': 'b',
                    'workerbuilddir': 'b',
                    'factory': self.factory,
                    'workernames': ['s1'],
                },
            )

    def test_init_workername_keyword(self):
        cfg = BuilderConfig(name='a b c', workername='a', factory=self.factory)
        self.assertEqual(cfg.workernames, ['a'])

    def test_init_workername_positional(self):
        with assertNotProducesWarnings(DeprecatedApiWarning):
            cfg = BuilderConfig('a b c', 'a', factory=self.factory)

        self.assertEqual(cfg.workernames, ['a'])

    def test_init_workernames_keyword(self):
        cfg = BuilderConfig(name='a b c', workernames=['a'], factory=self.factory)

        self.assertEqual(cfg.workernames, ['a'])

    def test_init_workernames_positional(self):
        with assertNotProducesWarnings(DeprecatedApiWarning):
            cfg = BuilderConfig('a b c', None, ['a'], factory=self.factory)

        self.assertEqual(cfg.workernames, ['a'])

    def test_init_workerbuilddir_keyword(self):
        cfg = BuilderConfig(
            name='a b c', workername='a', factory=self.factory, workerbuilddir="dir"
        )

        self.assertEqual(cfg.workerbuilddir, 'dir')

    def test_init_workerbuilddir_positional(self):
        with assertNotProducesWarnings(DeprecatedApiWarning):
            cfg = BuilderConfig('a b c', 'a', None, None, 'dir', factory=self.factory)

        self.assertEqual(cfg.workerbuilddir, 'dir')

    def test_init_next_worker_keyword(self):
        f = lambda: None
        cfg = BuilderConfig(name='a b c', workername='a', factory=self.factory, nextWorker=f)
        self.assertEqual(cfg.nextWorker, f)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import Any

from buildbot_worker import monkeypatches

# apply the same patches the worker does when it starts
monkeypatches.patch_all()


def add_debugging_monkeypatches() -> None:
    """
    DO NOT CALL THIS DIRECTLY

    This adds a few "harmless" monkeypatches which make it easier to debug
    failing tests.
    """
    from twisted.application.service import Service

    old_startService = Service.startService
    old_stopService = Service.stopService

    def startService(self: Service) -> None:
        assert not self.running
        return old_startService(self)

    def stopService(self: Service) -> None:
        assert self.running
        return old_stopService(self)

    Service.startService = startService  # type: ignore[method-assign]
    Service.stopService = stopService  # type: ignore[method-assign]


add_debugging_monkeypatches()

__all__: list[Any] = []

# import mock so we bail out early if it's not installed
try:
    from unittest import mock

    _ = mock
except ImportError:
    try:
        from unittest import mock
    except ImportError as e:
        raise ImportError("Buildbot tests require the 'mock' module; try 'pip install mock'") from e



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc[''], 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import copy
import datetime
import json
import types

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.changes import gerritchangesource
from buildbot.test import fakedb
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.fake.change import Change
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.runprocess import ExpectMasterShell
from buildbot.test.runprocess import MasterRunProcessMixin
from buildbot.test.util import changesource
from buildbot.test.util.state import StateTestMixin
from buildbot.util import datetime2epoch


class TestGerritHelpers(unittest.TestCase):
    def test_proper_json(self):
        self.assertEqual(
            "Justin Case <justin.case@example.com>",
            gerritchangesource._gerrit_user_to_author({
                "username": "justincase",
                "name": "Justin Case",
                "email": "justin.case@example.com",
            }),
        )

    def test_missing_username(self):
        self.assertEqual(
            "Justin Case <justin.case@example.com>",
            gerritchangesource._gerrit_user_to_author({
                "name": "Justin Case",
                "email": "justin.case@example.com",
            }),
        )

    def test_missing_name(self):
        self.assertEqual(
            "unknown <justin.case@example.com>",
            gerritchangesource._gerrit_user_to_author({"email": "justin.case@example.com"}),
        )
        self.assertEqual(
            "gerrit <justin.case@example.com>",
            gerritchangesource._gerrit_user_to_author(
                {"email": "justin.case@example.com"}, "gerrit"
            ),
        )
        self.assertEqual(
            "justincase <justin.case@example.com>",
            gerritchangesource._gerrit_user_to_author(
                {"username": "justincase", "email": "justin.case@example.com"}, "gerrit"
            ),
        )

    def test_missing_email(self):
        self.assertEqual(
            "Justin Case",
            gerritchangesource._gerrit_user_to_author({
                "username": "justincase",
                "name": "Justin Case",
            }),
        )
        self.assertEqual(
            "Justin Case", gerritchangesource._gerrit_user_to_author({"name": "Justin Case"})
        )
        self.assertEqual(
            "justincase", gerritchangesource._gerrit_user_to_author({"username": "justincase"})
        )
        self.assertEqual("unknown", gerritchangesource._gerrit_user_to_author({}))
        self.assertEqual("gerrit", gerritchangesource._gerrit_user_to_author({}, "gerrit"))


class TestGerritChangeSource(
    MasterRunProcessMixin,
    changesource.ChangeSourceMixin,
    StateTestMixin,
    TestReactorMixin,
    unittest.TestCase,
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_master_run_process()
        self._got_events = []
        yield self.setUpChangeSource()

    @defer.inlineCallbacks
    def create_gerrit(self, host, user, *args, **kwargs):
        http_url = kwargs.get("http_url", None)
        if http_url:
            self._http = yield fakehttpclientservice.HTTPClientService.getService(
                self.master, self, http_url + "/a", auth=kwargs.pop("expected_auth", None)
            )
        s = gerritchangesource.GerritChangeSource(host, user, *args, **kwargs)
        yield self.attachChangeSource(s)
        return s

    @defer.inlineCallbacks
    def create_gerrit_synchronized(self, host, user, *args, **kwargs):
        s = yield self.create_gerrit(host, user, *args, **kwargs)
        s._is_synchronized = True
        return s

    def assert_changes(self, expected_changes, ignore_keys):
        self.assertEqual(len(self.master.data.updates.changesAdded), len(expected_changes))
        for i, expected_change in enumerate(expected_changes):
            change = self.master.data.updates.changesAdded[i]
            for key in ignore_keys:
                del change[key]
            self.assertEqual(change, expected_change)

    def override_event_received(self, s):
        s.eventReceived = self._got_events.append

    def assert_events_received(self, events):
        self.assertEqual(self._got_events, events)

    # tests

    @defer.inlineCallbacks
    def test_describe(self):
        s = yield self.create_gerrit('somehost', 'someuser')
        self.assertSubstring("GerritChangeSource", s.describe())

    @defer.inlineCallbacks
    def test_name(self):
        s = yield self.create_gerrit('somehost', 'someuser')
        self.assertEqual("GerritChangeSource:someuser@somehost:29418", s.name)

        s = yield self.create_gerrit('somehost', 'someuser', name="MyName")
        self.assertEqual("MyName", s.name)

    patchset_created_event = {
        "uploader": {
            'name': 'uploader uploader',
            'email': 'uploader@example.com',
            'username': 'uploader',
        },
        "patchSet": {
            "number": 1,
            "revision": "29b73c3eb1aeaa9e6c7da520a940d60810e883db",
            "parents": ["7e563631188dcadf32aad0d8647c818834921a1e"],
            "ref": "refs/changes/21/4321/1",
            "uploader": {
                'name': 'uploader uploader',
                'email': 'uploader@example.com',
                'username': 'uploader',
            },
            "createdOn": 1627214047,
            "author": {
                'name': 'author author',
                'email': 'author@example.com',
                'username': 'author',
            },
            "kind": "REWORK",
            "sizeInsertions": 1,
            "sizeDeletions": 0,
        },
        "change": {
            "project": "test",
            "branch": "master",
            "id": "I21234123412341234123412341234",
            "number": 4321,
            "subject": "change subject",
            "owner": {'name': 'owner owner', 'email': 'owner@example.com', 'username': 'owner'},
            "url": "http://example.com/c/test/+/4321",
            "commitMessage": "test1\n\nChange-Id: I21234123412341234123412341234\n",
            "createdOn": 1627214047,
            "status": "NEW",
        },
        "project": "test",
        "refName": "refs/heads/master",
        "changeKey": {"id": "I21234123412341234123412341234"},
        "type": "patchset-created",
        "eventCreatedOn": 1627214048,
    }

    # this variable is reused in test_steps_source_repo
    # to ensure correct integration between change source and repo step
    expected_change_patchset_created = {
        'category': 'patchset-created',
        'files': ['unknown'],
        'repository': 'ssh://someuser@somehost:29418/test',
        'author': 'owner owner <owner@example.com>',
        'committer': None,
        'comments': 'change subject',
        'project': 'test',
        'branch': 'refs/changes/21/4321/1',
        'revision': '29b73c3eb1aeaa9e6c7da520a940d60810e883db',
        'codebase': None,
        'revlink': 'http://example.com/c/test/+/4321',
        'src': None,
        'when_timestamp': None,
    }

    @defer.inlineCallbacks
    def test_line_received_patchset_created(self):
        s = yield self.create_gerrit_synchronized('somehost', 'someuser')
        yield s._line_received_stream(json.dumps(self.patchset_created_event))

        self.assert_changes([self.expected_change_patchset_created], ignore_keys=['properties'])

    @defer.inlineCallbacks
    def test_line_received_patchset_created_props(self):
        s = yield self.create_gerrit_synchronized('somehost', 'someuser')
        yield s._line_received_stream(json.dumps(self.patchset_created_event))

        change = copy.deepcopy(self.expected_change_patchset_created)
        change['properties'] = {
            'event.change.branch': 'master',
            'event.change.commitMessage': 'test1\n\nChange-Id: I21234123412341234123412341234\n',
            'event.change.createdOn': 1627214047,
            'event.change.id': 'I21234123412341234123412341234',
            'event.change.number': 4321,
            'event.change.owner.email': 'owner@example.com',
            'event.change.owner.name': 'owner owner',
            'event.change.owner.username': 'owner',
            'event.change.project': 'test',
            'event.change.status': 'NEW',
            'event.change.subject': 'change subject',
            'event.change.url': 'http://example.com/c/test/+/4321',
            'event.changeKey.id': 'I21234123412341234123412341234',
            'event.patchSet.author.email': 'author@example.com',
            'event.patchSet.author.name': 'author author',
            'event.patchSet.author.username': 'author',
            'event.patchSet.createdOn': 1627214047,
            'event.patchSet.kind': 'REWORK',
            'event.patchSet.number': 1,
            'event.patchSet.parents': ['7e563631188dcadf32aad0d8647c818834921a1e'],
            'event.patchSet.ref': 'refs/changes/21/4321/1',
            'event.patchSet.revision': '29b73c3eb1aeaa9e6c7da520a940d60810e883db',
            'event.patchSet.sizeDeletions': 0,
            'event.patchSet.sizeInsertions': 1,
            'event.patchSet.uploader.email': 'uploader@example.com',
            'event.patchSet.uploader.name': 'uploader uploader',
            'event.patchSet.uploader.username': 'uploader',
            'event.project': 'test',
            'event.refName': 'refs/heads/master',
            'event.source': 'GerritChangeSource',
            'event.type': 'patchset-created',
            'event.uploader.email': 'uploader@example.com',
            'event.uploader.name': 'uploader uploader',
            'event.uploader.username': 'uploader',
            'target_branch': 'master',
        }
        self.assert_changes([change], ignore_keys=[])

    comment_added_event = {
        "type": "comment-added",
        "author": {'name': 'author author', 'email': 'author@example.com', 'username': 'author'},
        "approvals": [{"type": "Code-Review", "description": "Code-Review", "value": "0"}],
        "comment": "Patch Set 1:\n\ntest comment",
        "patchSet": {
            "number": 1,
            "revision": "29b73c3eb1aeaa9e6c7da520a940d60810e883db",
            "parents": ["7e563631188dcadf32aad0d8647c818834921a1e"],
            "ref": "refs/changes/21/4321/1",
            "uploader": {
                'name': 'uploader uploader',
                'email': 'uploader@example.com',
                'username': 'uploader',
            },
            "createdOn": 1627214047,
            "author": {
                'name': 'author author',
                'email': 'author@example.com',
                'username': 'author',
            },
            "kind": "REWORK",
            "sizeInsertions": 1,
            "sizeDeletions": 0,
        },
        "change": {
            "project": "test",
            "branch": "master",
            "id": "I21234123412341234123412341234",
            "number": 4321,
            "subject": "change subject",
            "owner": {'name': 'owner owner', 'email': 'owner@example.com', 'username': 'owner'},
            "url": "http://example.com/c/test/+/4321",
            "commitMessage": "test1\n\nChange-Id: I21234123412341234123412341234\n",
            "createdOn": 1627214047,
            "status": "NEW",
        },
        "project": "test",
        "refName": "refs/heads/master",
        "changeKey": {"id": "I21234123412341234123412341234"},
        "eventCreatedOn": 1627214102,
    }

    expected_change_comment_added = {
        'category': 'comment-added',
        'files': ['unknown'],
        'repository': 'ssh://someuser@somehost:29418/test',
        'author': 'owner owner <owner@example.com>',
        'committer': None,
        'comments': 'change subject',
        'project': 'test',
        'branch': 'refs/changes/21/4321/1',
        'revlink': 'http://example.com/c/test/+/4321',
        'codebase': None,
        'revision': '29b73c3eb1aeaa9e6c7da520a940d60810e883db',
        'src': None,
        'when_timestamp': None,
    }

    @defer.inlineCallbacks
    def test_line_received_comment_added(self):
        s = yield self.create_gerrit_synchronized(
            'somehost', 'someuser', handled_events=["comment-added"]
        )
        yield s._line_received_stream(json.dumps(self.comment_added_event))

        self.assert_changes([self.expected_change_comment_added], ignore_keys=['properties'])

    @defer.inlineCallbacks
    def test_line_received_ref_updated(self):
        s = yield self.create_gerrit_synchronized('somehost', 'someuser')
        yield s._line_received_stream(
            json.dumps({
                'type': 'ref-updated',
                'submitter': {
                    'name': 'tester',
                    'email': 'tester@example.com',
                    'username': 'tester',
                },
                'refUpdate': {
                    'oldRev': '12341234',
                    'newRev': '56785678',
                    'refName': 'refs/heads/master',
                    'project': 'test',
                },
                'eventCreatedOn': 1614528683,
            })
        )
        self.assertEqual(len(self.master.data.updates.changesAdded), 1)
        c = self.master.data.updates.changesAdded[0]
        self.assertEqual(
            c,
            {
                'files': ['unknown'],
                'comments': 'Gerrit: commit(s) pushed.',
                'author': 'tester <tester@example.com>',
                'committer': None,
                'revision': '56785678',
                'when_timestamp': None,
                'branch': 'master',
                'category': 'ref-updated',
                'revlink': '',
                'properties': {
                    'event.type': 'ref-updated',
                    'event.submitter.name': 'tester',
                    'event.submitter.email': 'tester@example.com',
                    'event.submitter.username': 'tester',
                    'event.refUpdate.oldRev': '12341234',
                    'event.refUpdate.newRev': '56785678',
                    'event.refUpdate.refName': 'refs/heads/master',
                    'event.refUpdate.project': 'test',
                    'event.source': 'GerritChangeSource',
                },
                'repository': 'ssh://someuser@somehost:29418/test',
                'codebase': None,
                'project': 'test',
                'src': None,
            },
        )

    @defer.inlineCallbacks
    def test_line_received_ref_updated_for_change(self):
        s = yield self.create_gerrit_synchronized('somehost', 'someuser')
        yield s._line_received_stream(
            json.dumps({
                'type': 'ref-updated',
                'submitter': {
                    'name': 'tester',
                    'email': 'tester@example.com',
                    'username': 'tester',
                },
                'refUpdate': {
                    'oldRev': '00000000',
                    'newRev': '56785678',
                    'refName': 'refs/changes/12/432112/1',
                    'project': 'test',
                },
                'eventCreatedOn': 1614528683,
            })
        )
        self.assertEqual(len(self.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def test_duplicate_non_source_events_not_ignored(self):
        s = yield self.create_gerrit_synchronized(
            'somehost',
            'someuser',
            handled_events=['patchset-created', 'ref-updated', 'change-merged', 'comment-added'],
        )
        yield s._line_received_stream(json.dumps(self.comment_added_event))
        self.assertEqual(len(self.master.data.updates.changesAdded), 1)

        yield s._line_received_stream(json.dumps(self.comment_added_event))
        self.assertEqual(len(self.master.data.updates.changesAdded), 2)

    @defer.inlineCallbacks
    def test_malformed_events_ignored(self):
        s = yield self.create_gerrit_synchronized('somehost', 'someuser')
        # "change" not in event
        yield s._line_received_stream(
            json.dumps({
                "type": "patchset-created",
                "patchSet": {"revision": 'abcdef', "number": '12'},
                'eventCreatedOn': 1614528683,
            })
        )
        self.assertEqual(len(self.master.data.updates.changesAdded), 0)

        # "patchSet" not in event
        yield s._line_received_stream(
            json.dumps({
                "type": "patchset-created",
                "change": {
                    "branch": "br",
                    # Note that this time "project" is a dictionary
                    "project": {"name": 'pr'},
                    "number": "4321",
                    "owner": {"name": 'Dustin', "email": 'dustin@mozilla.com'},
                    "url": "http://buildbot.net",
                    "subject": "fix 1234",
                },
                'eventCreatedOn': 1614528683,
            })
        )
        self.assertEqual(len(self.master.data.updates.changesAdded), 0)

    change_merged_event = {
        "type": "change-merged",
        "change": {
            "branch": "br",
            "project": "pr",
            "number": "4321",
            "owner": {"name": "Chuck", "email": "chuck@norris.com"},
            "url": "http://buildbot.net",
            "subject": "fix 1234",
        },
        "patchSet": {"revision": "abcdefj", "number": "13"},
        'eventCreatedOn': 1614528683,
    }

    @defer.inlineCallbacks
    def test_handled_events_filter_true(self):
        s = yield self.create_gerrit_synchronized(
            'somehost', 'some_choosy_user', handled_events=["change-merged"]
        )
        yield s._line_received_stream(json.dumps(self.change_merged_event))

        self.assertEqual(len(self.master.data.updates.changesAdded), 1)
        c = self.master.data.updates.changesAdded[0]
        self.assertEqual(c["category"], "change-merged")
        self.assertEqual(c["branch"], "br")

    @defer.inlineCallbacks
    def test_handled_events_filter_false(self):
        s = yield self.create_gerrit_synchronized('somehost', 'some_choosy_user')
        yield s._line_received_stream(json.dumps(self.change_merged_event))
        self.assertEqual(len(self.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def test_custom_handler(self):
        s = yield self.create_gerrit_synchronized(
            'somehost', 'some_choosy_user', handled_events=["change-merged"]
        )

        def custom_handler(self, properties, event):
            event['change']['project'] = "world"
            return self.addChangeFromEvent(properties, event)

        # Patches class to not bother with the inheritance
        s.eventReceived_change_merged = types.MethodType(custom_handler, s)
        yield s._line_received_stream(json.dumps(self.change_merged_event))

        self.assertEqual(len(self.master.data.updates.changesAdded), 1)
        c = self.master.data.updates.changesAdded[0]
        self.assertEqual(c['project'], "world")

    somehost_someuser_ssh_args = [
        "ssh",
        "-o",
        "BatchMode=yes",
        "-o",
        "ServerAliveInterval=15",
        "-o",
        "ServerAliveCountMax=3",
        "someuser@somehost",
        "-p",
        "29418",
        "gerrit",
        "stream-events",
    ]

    @defer.inlineCallbacks
    def test_activate(self):
        s = yield self.create_gerrit("somehost", "someuser", debug=True)

        self.reactor.expect_spawn("ssh", self.somehost_someuser_ssh_args)

        yield self.startChangeSource()
        s.activate()

        self.reactor.process_send_stderr(0, b"test stderr\n")
        self.reactor.process_send_stdout(0, b'{"type":"dropped-output", "eventCreatedOn": 123}\n')

        self.reactor.expect_process_signalProcess(0, "KILL")
        d = self.stopChangeSource()
        self.reactor.process_done(0, None)
        yield d

    @defer.inlineCallbacks
    def test_failure_backoff(self):
        s = yield self.create_gerrit("somehost", "someuser", debug=True)

        self.reactor.expect_spawn("ssh", self.somehost_someuser_ssh_args)

        yield self.startChangeSource()
        s.activate()

        pid = 0
        self.reactor.process_done(pid, None)
        pid += 1

        # The check happens as follows:
        #  - Advance reactor to just before required time (time - 0.05)
        #  - setup expectation for spawnProcess at that moment which ensures that spawnProcess was
        #    not called earlier,
        #  - Advance past the timeout and kill process, which ensures that the process has been
        #    created.
        self.reactor.advance(0.05)
        for time in [0.5, 0.5 * 1.5, 0.5 * 1.5 * 1.5, 0.5 * 1.5 * 1.5 * 1.5]:
            self.reactor.advance(time - 0.1)
            self.reactor.expect_spawn("ssh", self.somehost_someuser_ssh_args)
            self.reactor.advance(0.1)
            self.reactor.process_done(pid, None)
            pid += 1
            self.reactor.advance(0.05)

        yield self.stopChangeSource()

    def _build_messages_to_bytes(self, timestamps):
        messages = [
            json.dumps({
                "type": "patchset-created",
                'eventCreatedOn': timestamp,
            })
            for timestamp in timestamps
        ]
        out = b""
        for message in messages:
            out += (message + "\n").encode("utf-8")
        return out

    def _set_time_to(self, year, month, day, hours, minutes, seconds):
        self.reactor.advance(
            datetime2epoch(datetime.datetime(year, month, day, hours, minutes, seconds))
            - self.reactor.seconds()
        )

    @parameterized.expand([
        ("has_ts_in_db", True),
        ("has_no_ts_in_db", False),
    ])
    @defer.inlineCallbacks
    def test_poll_after_broken_connection(self, name, has_ts_in_db):
        self._set_time_to(2021, 1, 2, 3, 4, 5)
        start_time = self.reactor.seconds()

        s = yield self.create_gerrit(
            "somehost",
            "someuser",
            expected_auth=("user", "pass"),
            http_url="http://somehost",
            http_auth=("user", "pass"),
            http_poll_interval=30,
        )

        if has_ts_in_db:
            # Cannot use set_fake_state because the timestamp has already been retrieved
            s._last_event_ts = start_time - 124
            first_time_str = "2021-01-02 03:02:01"
        else:
            first_time_str = "2021-01-02 03:04:05"

        self.override_event_received(s)

        self.reactor.expect_spawn("ssh", self.somehost_someuser_ssh_args)
        self.reactor.expect_process_signalProcess(0, "KILL")

        # Initial poll
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": first_time_str},
            content=b"",
            processing_delay_s=1,
        )

        yield self.startChangeSource()
        s.activate()

        # Poll after timeout
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": first_time_str},
            content=self._build_messages_to_bytes([
                start_time + 1,
                start_time + 3,
                start_time + 5,
            ]),
        )
        self.reactor.advance(40)

        self.reactor.expect_spawn("ssh", self.somehost_someuser_ssh_args)
        self.reactor.expect_process_signalProcess(1, "KILL")
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:10"},
            content=self._build_messages_to_bytes([]),
        )

        # This is what triggers process startup
        self.reactor.process_done(0, None)

        self.reactor.advance(2)

        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:10"},
            content=self._build_messages_to_bytes([]),
        )
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:10"},
            content=self._build_messages_to_bytes([
                start_time + 41,
                start_time + 42,
                start_time + 43,  # ignored because data source switched to stream events
            ]),
        )

        self.reactor.process_send_stdout(
            1,
            self._build_messages_to_bytes([
                start_time + 41,
                start_time + 42,
            ]),
        )

        self.assertTrue(s._is_synchronized)

        d = self.stopChangeSource()
        self.reactor.process_done(1, None)
        yield d

        yield self.assert_state(s._oid, last_event_ts=start_time + 42)
        yield self.assert_state(
            s._oid, last_event_hashes=['f075e0927cab81dabee661a5aa3c65d502103a71']
        )

        self.assert_events_received([
            {'eventCreatedOn': start_time + 1, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 3, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 5, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 41, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 42, 'type': 'patchset-created'},
        ])

    @defer.inlineCallbacks
    def test_poll_after_broken_connection_with_message_before(self):
        self._set_time_to(2021, 1, 2, 3, 4, 5)
        start_time = self.reactor.seconds()

        s = yield self.create_gerrit(
            "somehost",
            "someuser",
            expected_auth=("user", "pass"),
            http_url="http://somehost",
            http_auth=("user", "pass"),
            http_poll_interval=30,
        )

        self.override_event_received(s)

        self.reactor.expect_spawn("ssh", self.somehost_someuser_ssh_args)
        self.reactor.expect_process_signalProcess(0, "KILL")

        # Initial poll
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:05"},
            content=b"",
        )

        yield self.startChangeSource()
        s.activate()

        self.reactor.advance(2)

        # Poll after messages below
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:05"},
            content=self._build_messages_to_bytes([]),
        )
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:05"},
            content=self._build_messages_to_bytes([]),
        )
        self.reactor.process_send_stdout(
            0,
            self._build_messages_to_bytes([
                start_time + 1,
                start_time + 2,
            ]),
        )

        # Poll after timeout
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:05"},
            content=self._build_messages_to_bytes([
                start_time + 1,
                start_time + 2,
            ]),
        )

        self.reactor.advance(40)

        self.reactor.expect_spawn("ssh", self.somehost_someuser_ssh_args)
        self.reactor.expect_process_signalProcess(1, "KILL")
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:07"},
            content=self._build_messages_to_bytes([]),
        )

        # This is what triggers process startup above
        self.reactor.process_done(0, None)

        # Stream should not be made primary until there are messages flowing
        self.assertFalse(s._is_synchronized)

        self.reactor.advance(2)

        # Poll after messages below
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:07"},
            content=self._build_messages_to_bytes([]),
        )
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:07"},
            content=self._build_messages_to_bytes([
                start_time + 41,
                start_time + 42,
                start_time + 43,  # ignored because data source switched to stream events
            ]),
        )

        self.reactor.process_send_stdout(
            1,
            self._build_messages_to_bytes([
                start_time + 41,
                start_time + 42,
            ]),
        )

        self.assertTrue(s._is_synchronized)

        d = self.stopChangeSource()
        self.reactor.process_done(1, None)
        yield d

        yield self.assert_state(s._oid, last_event_ts=start_time + 42)
        yield self.assert_state(
            s._oid, last_event_hashes=["f075e0927cab81dabee661a5aa3c65d502103a71"]
        )
        self.assert_events_received([
            {'eventCreatedOn': start_time + 1, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 2, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 41, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 42, 'type': 'patchset-created'},
        ])

    @defer.inlineCallbacks
    def test_poll_after_working_connection_but_no_messages(self):
        self._set_time_to(2021, 1, 2, 3, 4, 5)
        start_time = self.reactor.seconds()

        s = yield self.create_gerrit(
            "somehost",
            "someuser",
            expected_auth=("user", "pass"),
            http_url="http://somehost",
            http_auth=("user", "pass"),
            http_poll_interval=30,
        )

        self.override_event_received(s)

        self.reactor.expect_spawn("ssh", self.somehost_someuser_ssh_args)
        self.reactor.expect_process_signalProcess(0, "KILL")

        # Initial poll
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:05"},
            content=b"",
        )

        yield self.startChangeSource()
        s.activate()

        self.reactor.advance(2)

        # Poll after messages below
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:05"},
            content=self._build_messages_to_bytes([]),
        )
        self._http.expect(
            "get",
            "/plugins/events-log/events/",
            params={"t1": "2021-01-02 03:04:05"},
            content=self._build_messages_to_bytes([
                start_time + 1,
                start_time + 2,
            ]),
        )
        self.reactor.process_send_stdout(
            0,
            self._build_messages_to_bytes([
                start_time + 1,
                start_time + 2,
            ]),
        )

        self.assertTrue(s._is_synchronized)

        for _ in range(3):
            # Poll after timeout
            self._http.expect(
                "get",
                "/plugins/events-log/events/",
                params={"t1": "2021-01-02 03:04:07"},
                content=self._build_messages_to_bytes([]),
            )
            self.reactor.advance(40)

        self.reactor.advance(2)
        self.reactor.process_send_stdout(
            0,
            self._build_messages_to_bytes([
                start_time + 125,
                start_time + 126,
            ]),
        )

        for _ in range(3):
            # Poll after timeout
            self._http.expect(
                "get",
                "/plugins/events-log/events/",
                params={"t1": "2021-01-02 03:06:11"},
                content=self._build_messages_to_bytes([]),
            )
            self.reactor.advance(40)

        self.reactor.advance(2)
        self.reactor.process_send_stdout(
            0,
            self._build_messages_to_bytes([
                start_time + 256,
                start_time + 257,
            ]),
        )

        self.assertTrue(s._is_synchronized)

        d = self.stopChangeSource()
        self.reactor.process_done(0, None)
        yield d

        yield self.assert_state(s._oid, last_event_ts=start_time + 257)
        self.assert_events_received([
            {'eventCreatedOn': start_time + 1, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 2, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 125, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 126, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 256, 'type': 'patchset-created'},
            {'eventCreatedOn': start_time + 257, 'type': 'patchset-created'},
        ])

    # -------------------------------------------------------------------------
    # Test data for getFiles()
    # -------------------------------------------------------------------------
    query_files_success_line1 = {
        "patchSets": [
            {
                "number": 1,
                "files": [
                    {"file": "/COMMIT_MSG", "type": "ADDED", "insertions": 13, "deletions": 0},
                ],
            },
            {
                "number": 13,
                "files": [
                    {"file": "/COMMIT_MSG", "type": "ADDED", "insertions": 13, "deletions": 0},
                    {"file": "file1", "type": "MODIFIED", "insertions": 7, "deletions": 0},
                    {"file": "file2", "type": "MODIFIED", "insertions": 2, "deletions": -2},
                ],
            },
        ]
    }

    query_files_success_line2 = {"type": "stats", "rowCount": 1}

    query_files_success = '\n'.join([
        json.dumps(query_files_success_line1),
        json.dumps(query_files_success_line2),
    ]).encode('utf8')

    query_files_failure = b'{"type":"stats","rowCount":0}'

    @defer.inlineCallbacks
    def test_getFiles(self):
        s = yield self.create_gerrit_synchronized('host', 'user', gerritport=2222)
        exp_argv = [
            "ssh",
            "-o",
            "BatchMode=yes",
            "-o",
            "ServerAliveInterval=15",
            "-o",
            "ServerAliveCountMax=3",
            "user@host",
            "-p",
            "2222",
            "gerrit",
            "query",
            "1000",
            "--format",
            "JSON",
            "--files",
            "--patch-sets",
        ]

        self.expect_commands(
            ExpectMasterShell(exp_argv).stdout(self.query_files_success),
            ExpectMasterShell(exp_argv).stdout(self.query_files_failure),
        )

        res = yield s.getFiles(1000, 13)
        self.assertEqual(set(res), {'/COMMIT_MSG', 'file1', 'file2'})

        res = yield s.getFiles(1000, 13)
        self.assertEqual(res, ['unknown'])

        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def test_getFilesFromEvent(self):
        self.expect_commands(
            ExpectMasterShell([
                "ssh",
                "-o",
                "BatchMode=yes",
                "-o",
                "ServerAliveInterval=15",
                "-o",
                "ServerAliveCountMax=3",
                "user@host",
                "-p",
                "29418",
                "gerrit",
                "query",
                "4321",
                "--format",
                "JSON",
                "--files",
                "--patch-sets",
            ]).stdout(self.query_files_success)
        )

        s = yield self.create_gerrit_synchronized(
            'host', 'user', get_files=True, handled_events=["change-merged"]
        )

        yield s._line_received_stream(json.dumps(self.change_merged_event))
        c = self.master.data.updates.changesAdded[0]
        self.assertEqual(set(c['files']), {'/COMMIT_MSG', 'file1', 'file2'})

        self.assert_all_commands_ran()


class TestGerritEventLogPoller(
    changesource.ChangeSourceMixin, StateTestMixin, TestReactorMixin, unittest.TestCase
):
    NOW_TIMESTAMP = 1479302598
    EVENT_TIMESTAMP = 1479302599
    NOW_FORMATTED = '2016-11-16 13:23:18'
    EVENT_FORMATTED = '2016-11-16 13:23:19'
    OBJECTID = 1234

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpChangeSource()
        yield self.master.startService()
        self.addCleanup(self.master.stopService)

    @defer.inlineCallbacks
    def newChangeSource(self, **kwargs):
        auth = kwargs.pop('auth', ('log', 'pass'))
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, 'gerrit', auth=auth
        )
        self.changesource = gerritchangesource.GerritEventLogPoller(
            'gerrit',
            auth=auth,
            gitBaseURL="ssh://someuser@somehost:29418",
            pollAtLaunch=False,
            **kwargs,
        )

    @defer.inlineCallbacks
    def startChangeSource(self):
        yield self.changesource.setServiceParent(self.master)
        yield self.attachChangeSource(self.changesource)

    @defer.inlineCallbacks
    def test_describe(self):
        # describe is not used yet in buildbot nine, but it can still be useful in the future, so
        # lets implement and test it
        yield self.newChangeSource()
        self.assertSubstring('GerritEventLogPoller', self.changesource.describe())

    @defer.inlineCallbacks
    def test_name(self):
        yield self.newChangeSource()
        self.assertEqual('GerritEventLogPoller:gerrit', self.changesource.name)

    @defer.inlineCallbacks
    def test_lineReceived_patchset_created(self):
        yield self.master.db.insert_test_data([
            fakedb.Object(
                id=self.OBJECTID,
                name='GerritEventLogPoller:gerrit',
                class_name='GerritEventLogPoller',
            )
        ])
        yield self.newChangeSource(get_files=True)
        thirty_days_ago = datetime.datetime.fromtimestamp(
            self.reactor.seconds(), datetime.timezone.utc
        ) - datetime.timedelta(days=30)
        self._http.expect(
            method='get',
            ep='/plugins/events-log/events/',
            params={'t1': thirty_days_ago.strftime("%Y-%m-%d %H:%M:%S")},
            content_json={
                "type": "patchset-created",
                "change": {
                    "branch": "master",
                    "project": "test",
                    "number": "4321",
                    "owner": {"name": 'owner owner', "email": 'owner@example.com'},
                    "url": "http://example.com/c/test/+/4321",
                    "subject": "change subject",
                },
                "eventCreatedOn": self.EVENT_TIMESTAMP,
                "patchSet": {
                    'revision': "29b73c3eb1aeaa9e6c7da520a940d60810e883db",
                    'number': "1",
                    'ref': 'refs/changes/21/4321/1',
                },
            },
        )

        self._http.expect(
            method='get',
            ep='/changes/4321/revisions/1/files/',
            content=self.change_revision_resp,
        )

        yield self.startChangeSource()
        yield self.changesource._connector.poll()

        self.assertEqual(len(self.master.data.updates.changesAdded), 1)

        c = self.master.data.updates.changesAdded[0]
        expected_change = dict(TestGerritChangeSource.expected_change_patchset_created)
        for k, v in c.items():
            if k in ('files', 'properties'):
                continue
            self.assertEqual(expected_change[k], v)
        yield self.assert_state(self.OBJECTID, last_event_ts=self.EVENT_TIMESTAMP)

        self.assertEqual(set(c['files']), {'/COMMIT_MSG', 'file1'})

        # do a second poll, it should ask for the next events
        self._http.expect(
            method='get',
            ep='/plugins/events-log/events/',
            params={'t1': self.EVENT_FORMATTED},
            content_json={
                "type": "patchset-created",
                "change": {
                    "branch": "br",
                    "project": "pr",
                    "number": "4321",
                    "owner": {"name": 'Dustin', "email": 'dustin@mozilla.com'},
                    "url": "http://buildbot.net",
                    "subject": "fix 1234",
                },
                "eventCreatedOn": self.EVENT_TIMESTAMP + 1,
                "patchSet": {
                    'revision': "29b73c3eb1aeaa9e6c7da520a940d60810e883db",
                    'number': "1",
                    'ref': 'refs/changes/21/4321/1',
                },
            },
        )

        self._http.expect(
            method='get',
            ep='/changes/4321/revisions/1/files/',
            content=self.change_revision_resp,
        )

        yield self.changesource._connector.poll()
        yield self.assert_state(self.OBJECTID, last_event_ts=self.EVENT_TIMESTAMP + 1)

    change_revision_dict = {
        '/COMMIT_MSG': {'status': 'A', 'lines_inserted': 9, 'size_delta': 1, 'size': 1},
        'file1': {'lines_inserted': 9, 'lines_deleted': 2, 'size_delta': 1, 'size': 1},
    }
    change_revision_resp = b')]}\n' + json.dumps(change_revision_dict).encode('utf8')

    @defer.inlineCallbacks
    def test_getFiles(self):
        yield self.newChangeSource(get_files=True)
        yield self.startChangeSource()

        self._http.expect(
            method='get',
            ep='/changes/100/revisions/1/files/',
            content=self.change_revision_resp,
        )

        files = yield self.changesource.getFiles(100, 1)
        self.assertEqual(set(files), {'/COMMIT_MSG', 'file1'})

    @defer.inlineCallbacks
    def test_getFiles_handle_error(self):
        yield self.newChangeSource(get_files=True)
        yield self.startChangeSource()

        self._http.expect(
            method='get',
            ep='/changes/100/revisions/1/files/',
            content=b')]}\n',  # more than one line expected
        )

        files = yield self.changesource.getFiles(100, 1)
        self.assertEqual(files, [])
        self.assertEqual(len(self.flushLoggedErrors()), 1)


class TestGerritChangeFilter(unittest.TestCase):
    def test_event_type(self):
        props = {
            'event.type': 'patchset-created',
            'event.change.branch': 'master',
        }

        ch = Change(**TestGerritChangeSource.expected_change_patchset_created, properties=props)
        f = gerritchangesource.GerritChangeFilter(branch=["master"], eventtype=["patchset-created"])
        self.assertTrue(f.filter_change(ch))
        f = gerritchangesource.GerritChangeFilter(branch="master2", eventtype=["patchset-created"])
        self.assertFalse(f.filter_change(ch))
        f = gerritchangesource.GerritChangeFilter(branch="master", eventtype="ref-updated")
        self.assertFalse(f.filter_change(ch))
        self.assertEqual(
            repr(f),
            '<GerritChangeFilter on event.type in [\'ref-updated\'] and '
            'event.change.branch in [\'master\']>',
        )

    def create_props(self, branch, event_type):
        return {
            'event.type': event_type,
            'event.change.branch': branch,
        }

    def test_event_type_re(self):
        f = gerritchangesource.GerritChangeFilter(eventtype_re="patchset-.*")
        self.assertTrue(
            f.filter_change(Change(properties=self.create_props("br", "patchset-created")))
        )
        self.assertFalse(f.filter_change(Change(properties=self.create_props("br", "ref-updated"))))

    def test_event_type_fn(self):
        f = gerritchangesource.GerritChangeFilter(eventtype_fn=lambda t: t == "patchset-created")
        self.assertTrue(
            f.filter_change(Change(properties=self.create_props("br", "patchset-created")))
        )
        self.assertFalse(f.filter_change(Change(properties=self.create_props("br", "ref-updated"))))
        self.assertEqual(repr(f), '<GerritChangeFilter on <lambda>(eventtype)>')

    def test_branch_fn(self):
        f = gerritchangesource.GerritChangeFilter(branch_fn=lambda t: t == "br0")
        self.assertTrue(
            f.filter_change(Change(properties=self.create_props("br0", "patchset-created")))
        )
        self.assertFalse(
            f.filter_change(Change(properties=self.create_props("br1", "ref-updated")))
        )
        self.assertEqual(repr(f), '<GerritChangeFilter on <lambda>(branch)>')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import re
import traceback
from typing import Any
from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from twisted.python.reflect import accumulateClassList

from buildbot import config
from buildbot.process.properties import Properties
from buildbot.reporters.mail import VALID_EMAIL_ADDR
from buildbot.schedulers import base
from buildbot.util import identifiers


class ValidationError(ValueError):
    pass


class CollectedValidationError(ValueError):
    def __init__(self, errors):
        self.errors = errors
        super().__init__("\n".join([k + ":" + v for k, v in errors.items()]))


class ValidationErrorCollector:
    def __init__(self):
        self.errors = {}

    @defer.inlineCallbacks
    def collectValidationErrors(self, name, fn, *args, **kwargs):
        res = None
        try:
            res = yield fn(*args, **kwargs)
        except CollectedValidationError as err:
            for error_name, e in err.errors.items():
                self.errors[error_name] = e
        except ValueError as e:
            self.errors[name] = str(e)
        return res

    def maybeRaiseCollectedErrors(self):
        errors = self.errors
        if errors:
            raise CollectedValidationError(errors)


DefaultField = object()  # sentinel object to signal default behavior


class BaseParameter:
    """
    BaseParameter provides a base implementation for property customization
    """

    spec_attributes = [
        "name",
        "fullName",
        "label",
        "tablabel",
        "type",
        "default",
        "required",
        "multiple",
        "regex",
        "hide",
        "maxsize",
        "autopopulate",
        "tooltip",
    ]
    name = ""
    parentName = None
    label = ""
    tablabel = ""
    type: str = ""
    default: Any = ""
    required = False
    multiple = False
    regex = None
    debug = True
    hide = False
    maxsize: int | None = None
    autopopulate = None
    tooltip = ""

    @property
    def fullName(self):
        """A full name, intended to uniquely identify a parameter"""
        # join with '_' if both are set (cannot put '.', because it is used as
        # **kwargs)
        if self.parentName and self.name:
            return self.parentName + '_' + self.name
        # otherwise just use the one that is set
        # (this allows empty name for "anonymous nests")
        return self.name or self.parentName

    def setParent(self, parent):
        self.parentName = parent.fullName if parent else None

    def __init__(self, name, label=None, tablabel=None, regex=None, **kw):
        """
        @param name: the name of the field, used during posting values
                     back to the scheduler. This is not necessarily a UI value,
                     and there may be restrictions on the characters allowed for
                     this value. For example, HTML would require this field to
                     avoid spaces and other punctuation ('-', '.', and '_' allowed)
        @type name: unicode

        @param label: (optional) the name of the field, used for UI display.
        @type label: unicode or None (to use 'name')

        @param regex: (optional) regex to validate the value with. Not used by
                      all subclasses
        @type regex: unicode or regex
        """

        if name in ["owner", "builderNames", "builderid"]:
            config.error(f"{name} cannot be used as a parameter name, because it is reserved")
        self.name = name
        self.label = name if label is None else label
        self.tablabel = self.label if tablabel is None else tablabel
        if regex:
            self.regex = re.compile(regex)
        if 'value' in kw:
            config.error(
                f"Use default='{kw['value']}' instead of value=... to give a "
                "default Parameter value"
            )
        # all other properties are generically passed via **kw
        self.__dict__.update(kw)

    def getFromKwargs(self, kwargs):
        """Simple customization point for child classes that do not need the other
        parameters supplied to updateFromKwargs. Return the value for the property
        named 'self.name'.

        The default implementation converts from a list of items, validates using
        the optional regex field and calls 'parse_from_args' for the final conversion.
        """
        args = kwargs.get(self.fullName, [])

        # delete white space for args
        for arg in args:
            if isinstance(arg, str) and not arg.strip():
                args.remove(arg)

        if not args:
            if self.required:
                raise ValidationError(f"'{self.label}' needs to be specified")
            if self.multiple:
                args = self.default
            else:
                args = [self.default]

        if self.regex:
            for arg in args:
                if not self.regex.match(arg):
                    raise ValidationError(
                        f"{self.label}:'{arg}' does not match pattern '{self.regex.pattern}'"
                    )
        if self.maxsize is not None:
            for arg in args:
                if len(arg) > self.maxsize:
                    raise ValidationError(f"{self.label}: is too large {len(arg)} > {self.maxsize}")

        try:
            arg = self.parse_from_args(args)
        except Exception as e:
            # an exception will just display an alert in the web UI
            # also log the exception
            if self.debug:
                traceback.print_exc()
            raise e
        if arg is None:
            raise ValidationError(f"need {self.fullName}: no default provided by config")
        return arg

    def updateFromKwargs(self, properties, kwargs, collector, **unused):
        """Primary entry point to turn 'kwargs' into 'properties'"""
        properties[self.name] = self.getFromKwargs(kwargs)

    def parse_from_args(self, l):
        """Secondary customization point, called from getFromKwargs to turn
        a validated value into a single property value"""
        if self.multiple:
            return [self.parse_from_arg(arg) for arg in l]
        return self.parse_from_arg(l[0])

    def parse_from_arg(self, s):
        return s

    def getSpec(self):
        spec_attributes = []
        accumulateClassList(self.__class__, 'spec_attributes', spec_attributes)
        ret = {}
        for i in spec_attributes:
            ret[i] = getattr(self, i)
        return ret


class FixedParameter(BaseParameter):
    """A fixed parameter that cannot be modified by the user."""

    type = "fixed"
    hide = True
    default = ""

    def parse_from_args(self, l):
        return self.default


class StringParameter(BaseParameter):
    """A simple string parameter"""

    spec_attributes = ["size"]
    type = "text"
    size = 10

    def parse_from_arg(self, s):
        return s


class TextParameter(StringParameter):
    """A generic string parameter that may span multiple lines"""

    spec_attributes = ["cols", "rows"]
    type = "textarea"
    cols = 80
    rows = 20

    def value_to_text(self, value):
        return str(value)


class IntParameter(StringParameter):
    """An integer parameter"""

    type = "int"
    default = 0
    parse_from_arg = int  # will throw an exception if parse fail


class BooleanParameter(BaseParameter):
    """A boolean parameter"""

    type = "bool"

    def getFromKwargs(self, kwargs):
        return kwargs.get(self.fullName, [self.default]) == [True]


class UserNameParameter(StringParameter):
    """A username parameter to supply the 'owner' of a build"""

    spec_attributes = ["need_email"]
    type = "username"
    default = ""
    size = 30
    need_email = True

    def __init__(self, name="username", label="Your name:", **kw):
        super().__init__(name, label, **kw)

    def parse_from_arg(self, s):
        if not s and not self.required:
            return s
        if self.need_email:
            res = VALID_EMAIL_ADDR.search(s)
            if res is None:
                raise ValidationError(
                    f"{self.name}: please fill in email address in the "
                    "form 'User <email@email.com>'"
                )
        return s


class ChoiceStringParameter(BaseParameter):
    """A list of strings, allowing the selection of one of the predefined values.
    The 'strict' parameter controls whether values outside the predefined list
    of choices are allowed"""

    spec_attributes = ["choices", "strict"]
    type = "list"
    choices: list[str] = []
    strict = True

    def parse_from_arg(self, s):
        if self.strict and s not in self.choices:
            raise ValidationError(
                f"'{s}' does not belong to list of available choices '{self.choices}'"
            )
        return s

    def getChoices(self, master, scheduler, buildername):
        return self.choices


class InheritBuildParameter(ChoiceStringParameter):
    """A parameter that takes its values from another build"""

    type = ChoiceStringParameter.type
    name = "inherit"
    compatible_builds = None

    def getChoices(self, master, scheduler, buildername):
        return self.compatible_builds(master, buildername)

    def getFromKwargs(self, kwargs):
        raise ValidationError("InheritBuildParameter can only be used by properties")

    def updateFromKwargs(self, master, properties, changes, kwargs, **unused):
        arg = kwargs.get(self.fullName, [""])[0]
        split_arg = arg.split(" ")[0].split("/")
        if len(split_arg) != 2:
            raise ValidationError(f"bad build: {arg}")
        builder_name, build_num = split_arg

        builder_dict = master.data.get(('builders', builder_name))
        if builder_dict is None:
            raise ValidationError(f"unknown builder: {builder_name} in {arg}")

        build_dict = master.data.get(
            ('builders', builder_name, 'builds', build_num), fields=['properties']
        )
        if build_dict is None:
            raise ValidationError(f"unknown build: {builder_name} in {arg}")

        props = {self.name: (arg.split(" ")[0])}
        for name, (value, source) in build_dict['properties']:
            if source == "Force Build Form":
                if name == "owner":
                    name = "orig_owner"
                props[name] = value
        properties.update(props)
        # FIXME: this does not do what we expect, but updateFromKwargs is not used either.
        # This needs revisiting when the build parameters are fixed:
        # changes.extend(b.changes)


class WorkerChoiceParameter(ChoiceStringParameter):
    """A parameter that lets the worker name be explicitly chosen.

    This parameter works in conjunction with 'buildbot.process.builder.enforceChosenWorker',
    which should be added as the 'canStartBuild' parameter to the Builder.

    The "anySentinel" parameter represents the sentinel value to specify that
    there is no worker preference.
    """

    anySentinel = '-any-'
    label = 'Worker'
    required = False
    strict = False

    def __init__(self, name='workername', **kwargs):
        super().__init__(name, **kwargs)

    def updateFromKwargs(self, kwargs, **unused):
        workername = self.getFromKwargs(kwargs)
        if workername == self.anySentinel:
            # no preference, so don't set a parameter at all
            return
        super().updateFromKwargs(kwargs=kwargs, **unused)

    @defer.inlineCallbacks
    def getChoices(self, master, scheduler, buildername):
        if buildername is None:
            # this is the "Force All Builds" page
            workers = yield self.master.data.get(('workers',))
        else:
            builder = yield self.master.data.get(('builders', buildername))
            workers = yield self.master.data.get(('builders', builder['builderid'], 'workers'))

        workernames = [worker['name'] for worker in workers]
        workernames.sort()
        workernames.insert(0, self.anySentinel)
        return workernames


class FileParameter(BaseParameter):
    """A parameter which allows to download a whole file and store it as a property or patch"""

    type = 'file'
    maxsize = 1024 * 1024 * 10  # 10M


class NestedParameter(BaseParameter):
    """A 'parent' parameter for a set of related parameters. This provides a
    logical grouping for the child parameters.

    Typically, the 'fullName' of the child parameters mix in the parent's
    'fullName'. This allows for a field to appear multiple times in a form
    (for example, two codebases each have a 'branch' field).

    If the 'name' of the parent is the empty string, then the parent's name
    does not mix in with the child 'fullName'. This is useful when a field
    will not appear multiple time in a scheduler but the logical grouping is
    helpful.

    The result of a NestedParameter is typically a dictionary, with the key/value
    being the name/value of the children.
    """

    spec_attributes = [
        "layout",
        "columns",
    ]  # field is recursive, and thus managed in custom getSpec
    type = 'nested'
    layout = 'vertical'
    fields = None
    columns: int | None = None

    def __init__(self, name, fields, **kwargs):
        super().__init__(fields=fields, name=name, **kwargs)
        # reasonable defaults for the number of columns
        if self.columns is None:
            num_visible_fields = len([field for field in fields if not field.hide])
            if num_visible_fields >= 4:
                self.columns = 2
            else:
                self.columns = 1
        if self.columns > 4:
            config.error("UI only support up to 4 columns in nested parameters")

        # fix up the child nodes with the parent (use None for now):
        self.setParent(None)

    def setParent(self, parent):
        super().setParent(parent)
        for field in self.fields:  # pylint: disable=not-an-iterable
            field.setParent(self)

    @defer.inlineCallbacks
    def collectChildProperties(self, kwargs, properties, collector, **kw):
        """Collapse the child values into a dictionary. This is intended to be
        called by child classes to fix up the fullName->name conversions."""

        childProperties = {}
        for field in self.fields:  # pylint: disable=not-an-iterable
            yield collector.collectValidationErrors(
                field.fullName,
                field.updateFromKwargs,
                kwargs=kwargs,
                properties=childProperties,
                collector=collector,
                **kw,
            )
        kwargs[self.fullName] = childProperties

    @defer.inlineCallbacks
    def updateFromKwargs(self, kwargs, properties, collector, **kw):
        """By default, the child values will be collapsed into a dictionary. If
        the parent is anonymous, this dictionary is the top-level properties."""
        yield self.collectChildProperties(
            kwargs=kwargs, properties=properties, collector=collector, **kw
        )
        # default behavior is to set a property
        #  -- use setdefault+update in order to collapse 'anonymous' nested
        #     parameters correctly
        if self.name:
            d = properties.setdefault(self.name, {})
        else:
            # if there's no name, collapse this nest all the way
            d = properties
        d.update(kwargs[self.fullName])

    def getSpec(self):
        ret = super().getSpec()
        # pylint: disable=not-an-iterable
        ret['fields'] = [field.getSpec() for field in self.fields]
        return ret


ParameterGroup = NestedParameter


class AnyPropertyParameter(NestedParameter):
    """A generic property parameter, where both the name and value of the property
    must be given."""

    type = NestedParameter.type

    def __init__(self, name, **kw):
        fields = [
            StringParameter(name='name', label="Name:"),
            StringParameter(name='value', label="Value:"),
        ]
        super().__init__(name, label='', fields=fields, **kw)

    def getFromKwargs(self, kwargs):
        raise ValidationError("AnyPropertyParameter can only be used by properties")

    @defer.inlineCallbacks
    def updateFromKwargs(self, master, properties, kwargs, collector, **kw):
        yield self.collectChildProperties(
            master=master, properties=properties, kwargs=kwargs, collector=collector, **kw
        )

        pname = kwargs[self.fullName].get("name", "")
        pvalue = kwargs[self.fullName].get("value", "")
        if not pname:
            return

        validation = master.config.validation
        pname_validate = validation['property_name']
        pval_validate = validation['property_value']

        if not pname_validate.match(pname) or not pval_validate.match(pvalue):
            raise ValidationError(f"bad property name='{pname}', value='{pvalue}'")
        properties[pname] = pvalue


class CodebaseParameter(NestedParameter):
    """A parameter whose result is a codebase specification instead of a property"""

    type = NestedParameter.type
    codebase = ''

    def __init__(
        self,
        codebase,
        name=None,
        label=None,
        branch=DefaultField,
        revision=DefaultField,
        repository=DefaultField,
        project=DefaultField,
        patch=None,
        **kwargs,
    ):
        """
        A set of properties that will be used to generate a codebase dictionary.

        The branch/revision/repository/project should each be a parameter that
        will map to the corresponding value in the sourcestamp. Use None to disable
        the field.

        @param codebase: name of the codebase; used as key for the sourcestamp set
        @type codebase: unicode

        @param name: optional override for the name-currying for the subfields
        @type codebase: unicode

        @param label: optional override for the label for this set of parameters
        @type codebase: unicode
        """

        name = name or codebase
        if label is None and codebase:
            label = "Codebase: " + codebase

        fields_dict = {
            "branch": branch,
            "revision": revision,
            "repository": repository,
            "project": project,
        }
        for k, v in fields_dict.items():
            if v is DefaultField:
                v = StringParameter(name=k, label=k.capitalize() + ":")
            elif isinstance(v, str):
                v = FixedParameter(name=k, default=v)
            fields_dict[k] = v

        fields = [val for k, val in sorted(fields_dict.items(), key=lambda x: x[0]) if val]
        if patch is not None:
            if patch.name != "patch":
                config.error("patch parameter of a codebase must be named 'patch'")
            fields.append(patch)
            if self.columns is None and 'columns' not in kwargs:
                self.columns = 1

        super().__init__(name=name, label=label, codebase=codebase, fields=fields, **kwargs)

    def createSourcestamp(self, properties, kwargs):
        # default, just return the things we put together
        return kwargs.get(self.fullName, {})

    @defer.inlineCallbacks
    def updateFromKwargs(self, sourcestamps, kwargs, properties, collector, **kw):
        yield self.collectChildProperties(
            sourcestamps=sourcestamps,
            properties=properties,
            kwargs=kwargs,
            collector=collector,
            **kw,
        )

        # convert the "property" to a sourcestamp
        ss = self.createSourcestamp(properties, kwargs)
        if ss is not None:
            patch = ss.pop('patch', None)
            if patch is not None:
                for k, v in patch.items():
                    ss['patch_' + k] = v

            sourcestamps[self.codebase] = ss


def oneCodebase(**kw):
    return [CodebaseParameter('', **kw)]


class PatchParameter(NestedParameter):
    """A patch parameter contains pre-configure UI for all the needed components for a
    sourcestamp patch
    """

    columns = 1

    def __init__(self, **kwargs):
        name = kwargs.pop('name', 'patch')
        default_fields = [
            FileParameter('body'),
            IntParameter('level', default=1),
            StringParameter('author', default=""),
            StringParameter('comment', default=""),
            StringParameter('subdir', default="."),
        ]
        fields = [kwargs.pop(field.name, field) for field in default_fields]
        super().__init__(name, fields=fields, **kwargs)


class ForceScheduler(base.ReconfigurableBaseScheduler):
    """
    ForceScheduler implements the backend for a UI to allow customization of
    builds. For example, a web form be populated to trigger a build.
    """

    compare_attrs: ClassVar[Sequence[str]] = (
        *base.ReconfigurableBaseScheduler.compare_attrs,
        "builderNames",
        "reason",
        "username",
        "forcedProperties",
        "forced_priority",
    )

    def __init__(self, name, builderNames, **kwargs):
        super().__init__(name=name, builderNames=builderNames, **kwargs)

    def checkConfig(  # type: ignore[override]
        self,
        builderNames,
        username: UserNameParameter | None = None,
        reason: StringParameter | None = None,
        reasonString="A build was forced by '%(owner)s': %(reason)s",
        buttonName=None,
        codebases=None,
        label=None,
        properties=None,
        priority: IntParameter | None = None,
        **kwargs: Any,
    ):
        super().checkConfig(builderNames=builderNames, **kwargs)

        if not self.name:
            config.error(f"ForceScheduler name must not be empty: {self.name!r}")

        assert self.name is not None

        if not identifiers.ident_re.match(self.name):
            config.error(f"ForceScheduler name must be an identifier: {self.name!r}")

        if not self.checkIfListOfType(builderNames, (str,)):
            config.error(
                f"ForceScheduler '{self.name}': builderNames must be a list of strings: {builderNames!r}"
            )

        if username is not None and not self.checkIfType(username, BaseParameter):
            config.error(
                f"ForceScheduler '{self.name}': username must be a StringParameter: {username!r}"
            )

        if reason is not None and not self.checkIfType(reason, BaseParameter):
            config.error(
                f"ForceScheduler '{self.name}': reason must be a StringParameter: {reason!r}"
            )

        if properties is not None and not self.checkIfListOfType(properties, BaseParameter):
            config.error(
                f"ForceScheduler '{self.name}': properties must be "
                f"a list of BaseParameters: {properties!r}"
            )

        if priority is not None and not self.checkIfType(priority, IntParameter):
            config.error(
                f"ForceScheduler '{self.name}': priority must be a IntParameter: {priority!r}"
            )

        if codebases is not None:
            if not codebases:
                config.error(
                    f"ForceScheduler '{self.name}': 'codebases' cannot be empty;"
                    f" use [CodebaseParameter(codebase='', hide=True)] if needed: "
                    f"{codebases!r} "
                )
            if not isinstance(codebases, list):
                config.error(
                    f"ForceScheduler '{self.name}': 'codebases' should be a list of strings "
                    f"or CodebaseParameter, not {type(codebases)}"
                )

            for codebase in codebases:
                if not isinstance(codebase, (str, CodebaseParameter)):
                    config.error(
                        f"ForceScheduler '{self.name}': 'codebases' must be a list of strings "
                        f"or CodebaseParameter objects: {codebases!r}"
                    )

    @defer.inlineCallbacks
    def reconfigService(  # type: ignore[override]
        self,
        builderNames,
        username: UserNameParameter | None = None,
        reason: StringParameter | None = None,
        reasonString="A build was forced by '%(owner)s': %(reason)s",
        buttonName=None,
        codebases=None,
        label=None,
        properties=None,
        priority: IntParameter | None = None,
        **kwargs: Any,
    ):
        if reason is None:
            reason = StringParameter(name="reason", default="force build", size=20)
        self.reason = reason

        if username is None:
            username = UserNameParameter()
        self.username = username

        if priority is None:
            priority = IntParameter(name="priority", default=0)
        self.forced_priority = priority

        self.label = self.name if label is None else label

        self.forcedProperties = []
        if codebases is None:
            codebases = [CodebaseParameter(codebase='')]

        codebase_dict = {}
        for codebase in codebases:
            if isinstance(codebase, str):
                codebase = CodebaseParameter(codebase=codebase)

            self.forcedProperties.append(codebase)
            codebase_dict[codebase.codebase] = {"branch": '', "repository": '', "revision": ''}

        yield super().reconfigService(builderNames=builderNames, codebases=codebase_dict, **kwargs)

        if properties:
            self.forcedProperties.extend(properties)

        # this is used to simplify the template
        self.all_fields = [NestedParameter(name='', fields=[username, reason, priority])]
        self.all_fields.extend(self.forcedProperties)

        self.reasonString = reasonString
        self.buttonName = buttonName or self.name

    def checkIfType(self, obj, chkType):
        return isinstance(obj, chkType)

    def checkIfListOfType(self, obj, chkType):
        isListOfType = True

        if self.checkIfType(obj, list):
            for item in obj:
                if not self.checkIfType(item, chkType):
                    isListOfType = False
                    break
        else:
            isListOfType = False

        return isListOfType

    @defer.inlineCallbacks
    def gatherPropertiesAndChanges(self, collector, **kwargs):
        properties = {}
        changeids = []
        sourcestamps = {}

        for param in self.forcedProperties:
            yield collector.collectValidationErrors(
                param.fullName,
                param.updateFromKwargs,
                master=self.master,
                properties=properties,
                changes=changeids,
                sourcestamps=sourcestamps,
                collector=collector,
                kwargs=kwargs,
            )
        changeids = [(type(a) == int and a) or a.number for a in changeids]

        real_properties = Properties()
        for pname, pvalue in properties.items():
            real_properties.setProperty(pname, pvalue, "Force Build Form")

        return (real_properties, changeids, sourcestamps)

    @defer.inlineCallbacks
    def computeBuilderNames(self, builderNames=None, builderid=None):
        if builderNames is None:
            if builderid is not None:
                builder = yield self.master.data.get(('builders', str(builderid)))
                builderNames = [builder['name']]
            else:
                builderNames = self.builderNames
        else:
            builderNames = sorted(set(builderNames).intersection(self.builderNames))
        return builderNames

    @defer.inlineCallbacks
    def force(self, owner, builderNames=None, builderid=None, **kwargs):
        """
        We check the parameters, and launch the build, if everything is correct
        """
        builderNames = yield self.computeBuilderNames(builderNames, builderid)
        if not builderNames:
            raise KeyError("builderNames not specified or not supported")

        # Currently the validation code expects all kwargs to be lists
        # I don't want to refactor that now so much sure we comply...
        kwargs = dict((k, [v]) if not isinstance(v, list) else (k, v) for k, v in kwargs.items())

        # probably need to clean that out later as the IProperty is already a
        # validation mechanism
        collector = ValidationErrorCollector()
        reason = yield collector.collectValidationErrors(
            self.reason.fullName, self.reason.getFromKwargs, kwargs
        )
        if owner is None or owner == "anonymous":
            owner = yield collector.collectValidationErrors(
                self.username.fullName, self.username.getFromKwargs, kwargs
            )

        priority = yield collector.collectValidationErrors(
            self.forced_priority.fullName, self.forced_priority.getFromKwargs, kwargs
        )

        properties, _, sourcestamps = yield self.gatherPropertiesAndChanges(collector, **kwargs)

        collector.maybeRaiseCollectedErrors()

        properties.setProperty("reason", reason, "Force Build Form")
        properties.setProperty("owner", owner, "Force Build Form")

        r = self.reasonString % {'owner': owner, 'reason': reason}

        # turn sourcestamps into a list
        for cb, ss in sourcestamps.items():
            ss['codebase'] = cb
        sourcestamps = list(sourcestamps.values())

        # everything is validated, we can create our source stamp, and
        # buildrequest
        res = yield self.addBuildsetForSourceStampsWithDefaults(
            reason=r,
            sourcestamps=sourcestamps,
            properties=properties,
            builderNames=builderNames,
            priority=priority,
        )

        return res



#!/usr/bin/env jython


import glob
import sys

testdir = sys.argv[1]

orderfiles = glob.glob(testdir + '/*.tests')

# wee. just be glad I didn't make this one gigantic nested listcomp.
# anyway, this builds a once-nested list of files to test.

# open!
files = [open(fn) for fn in orderfiles]

# create prelim list of lists of files!
files = [f.readlines() for f in files]

# shwack newlines and filter out empties!
files = [filter(None, [fn.strip() for fn in fs]) for fs in files]

# prefix with testdir
files = [[testdir + '/' + fn.strip() for fn in fs] for fs in files]

print("Will run these tests:", files)

i = 0

for testlist in files:
    print("===========================")
    print("running tests from testlist", orderfiles[i])
    print("---------------------------")
    i = i + 1

    for test in testlist:
        print("running test", test)

        try:
            with open(test) as f:
                exec(f.read(), globals().copy())

        except Exception:
            ei = sys.exc_info()
            print("TEST FAILURE:", ei[1])

        else:
            print("SUCCESS")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import www
from buildbot.www import authz
from buildbot.www.authz.endpointmatchers import AnyControlEndpointMatcher
from buildbot.www.authz.endpointmatchers import AnyEndpointMatcher
from buildbot.www.authz.endpointmatchers import BranchEndpointMatcher
from buildbot.www.authz.endpointmatchers import EndpointMatcherBase
from buildbot.www.authz.endpointmatchers import ForceBuildEndpointMatcher
from buildbot.www.authz.endpointmatchers import RebuildBuildEndpointMatcher
from buildbot.www.authz.endpointmatchers import StopBuildEndpointMatcher
from buildbot.www.authz.endpointmatchers import ViewBuildsEndpointMatcher
from buildbot.www.authz.roles import RolesFromDomain
from buildbot.www.authz.roles import RolesFromEmails
from buildbot.www.authz.roles import RolesFromGroups
from buildbot.www.authz.roles import RolesFromOwner

if TYPE_CHECKING:
    from buildbot.util.twisted import InlineCallbacksType


class Authz(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self) -> InlineCallbacksType[None]:  # type: ignore[override]
        self.setup_test_reactor()
        authzcfg = authz.Authz(
            # simple matcher with '*' glob character
            stringsMatcher=authz.fnmatchStrMatcher,
            # stringsMatcher = authz.Authz.reStrMatcher,  # if you prefer
            # regular expressions
            allowRules=[
                # admins can do anything,
                # defaultDeny=False: if user does not have the admin role, we
                # continue parsing rules
                AnyEndpointMatcher(role="admins", defaultDeny=False),
                # rules for viewing builds, builders, step logs
                # depending on the sourcestamp or buildername
                ViewBuildsEndpointMatcher(branch="secretbranch", role="agents"),
                ViewBuildsEndpointMatcher(project="secretproject", role="agents"),
                ViewBuildsEndpointMatcher(branch="*", role="*"),
                ViewBuildsEndpointMatcher(project="*", role="*"),
                StopBuildEndpointMatcher(role="owner"),
                RebuildBuildEndpointMatcher(role="owner"),
                # nine-* groups can do stuff on the nine branch
                BranchEndpointMatcher(branch="nine", role="nine-*"),
                # eight-* groups can do stuff on the eight branch
                BranchEndpointMatcher(branch="eight", role="eight-*"),
                # *-try groups can start "try" builds
                ForceBuildEndpointMatcher(builder="try", role="*-developers"),
                # *-mergers groups can start "merge" builds
                ForceBuildEndpointMatcher(builder="merge", role="*-mergers"),
                # *-releasers groups can start "release" builds
                ForceBuildEndpointMatcher(builder="release", role="*-releasers"),
                # finally deny any control endpoint for non-admin users
                AnyControlEndpointMatcher(role="admins"),
            ],
            roleMatchers=[
                RolesFromGroups(groupPrefix="buildbot-"),
                RolesFromEmails(admins=["homer@springfieldplant.com"], agents=["007@mi6.uk"]),
                RolesFromOwner(role="owner"),
                RolesFromDomain(admins=["mi7.uk"]),
            ],
        )
        self.users = {
            "homer": {"email": "homer@springfieldplant.com"},
            "bond": {"email": "007@mi6.uk"},
            "moneypenny": {"email": "moneypenny@mi7.uk"},
            "nineuser": {
                "email": "user@nine.com",
                "groups": ["buildbot-nine-mergers", "buildbot-nine-developers"],
            },
            "eightuser": {"email": "user@eight.com", "groups": ["buildbot-eight-developers"]},
        }
        self.master = yield self.make_master(url='h:/a/b/', authz=authzcfg)
        self.authz = self.master.authz
        yield self.master.db.insert_test_data([
            fakedb.Builder(id=77, name="mybuilder"),
            fakedb.Master(id=88),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Buildset(id=8822),
            fakedb.BuildsetProperty(
                buildsetid=8822, property_name='owner', property_value='["user@nine.com", "force"]'
            ),
            fakedb.BuildRequest(id=82, buildsetid=8822, builderid=77),
            fakedb.Build(
                id=13, builderid=77, masterid=88, workerid=13, buildrequestid=82, number=3
            ),
            fakedb.Build(
                id=14, builderid=77, masterid=88, workerid=13, buildrequestid=82, number=4
            ),
            fakedb.Build(
                id=15, builderid=77, masterid=88, workerid=13, buildrequestid=82, number=5
            ),
        ])

    def setAllowRules(self, allow_rules: list[EndpointMatcherBase]) -> None:
        # we should add links to authz and master instances in each new rule
        for r in allow_rules:
            r.setAuthz(self.authz)

        self.authz.allowRules = allow_rules

    def assertUserAllowed(self, ep: str, action: str, options: dict, user: str) -> defer.Deferred:
        return self.authz.assertUserAllowed(tuple(ep.split("/")), action, options, self.users[user])

    @defer.inlineCallbacks
    def assertUserForbidden(
        self, ep: str, action: str, options: dict, user: str
    ) -> InlineCallbacksType[None]:
        try:
            yield self.authz.assertUserAllowed(
                tuple(ep.split("/")), action, options, self.users[user]
            )
        except authz.Forbidden as err:
            self.assertIn('need to have role', repr(err))
        else:
            self.fail('authz.Forbidden with error "need to have role" was expected!')

    @defer.inlineCallbacks
    def test_anyEndpoint(self) -> InlineCallbacksType[None]:
        # admin users can do anything
        yield self.assertUserAllowed("foo/bar", "get", {}, "homer")
        yield self.assertUserAllowed("foo/bar", "stop", {}, "moneypenny")
        # non-admin user can only do "get" action
        yield self.assertUserAllowed("foo/bar", "get", {}, "bond")
        # non-admin user cannot do control actions
        yield self.assertUserForbidden("foo/bar", "stop", {}, "bond")

        # non-admin user cannot do any actions
        allow_rules: list[EndpointMatcherBase] = [
            AnyEndpointMatcher(role="admins"),
        ]
        self.setAllowRules(allow_rules)
        yield self.assertUserForbidden("foo/bar", "get", {}, "bond")
        yield self.assertUserForbidden("foo/bar", "stop", {}, "bond")

    @defer.inlineCallbacks
    def test_stopBuild(self) -> InlineCallbacksType[None]:
        # admin can always stop
        yield self.assertUserAllowed("builds/13", "stop", {}, "homer")
        # owner can always stop
        yield self.assertUserAllowed("builds/13", "stop", {}, "nineuser")
        yield self.assertUserAllowed("buildrequests/82", "stop", {}, "nineuser")
        # not owner cannot stop
        yield self.assertUserForbidden("builds/13", "stop", {}, "eightuser")
        yield self.assertUserForbidden("buildrequests/82", "stop", {}, "eightuser")

        # can stop build/buildrequest with matching builder
        allow_rules = [
            StopBuildEndpointMatcher(role="eight-*", builder="mybuilder"),
            AnyEndpointMatcher(role="admins"),
        ]
        self.setAllowRules(allow_rules)
        yield self.assertUserAllowed("builds/13", "stop", {}, "eightuser")
        yield self.assertUserAllowed("buildrequests/82", "stop", {}, "eightuser")
        yield self.assertUserForbidden("builds/999", "stop", {}, "eightuser")
        yield self.assertUserForbidden("buildrequests/999", "stop", {}, "eightuser")

        # cannot stop build/buildrequest with non-matching builder
        allow_rules = [
            StopBuildEndpointMatcher(role="eight-*", builder="foo"),
            AnyEndpointMatcher(role="admins"),
        ]
        self.setAllowRules(allow_rules)
        yield self.assertUserForbidden("builds/13", "stop", {}, "eightuser")
        yield self.assertUserForbidden("buildrequests/82", "stop", {}, "eightuser")

    @defer.inlineCallbacks
    def test_rebuildBuild(self) -> InlineCallbacksType[None]:
        # admin can rebuild
        yield self.assertUserAllowed("builds/13", "rebuild", {}, "homer")
        # owner can always rebuild
        yield self.assertUserAllowed("builds/13", "rebuild", {}, "nineuser")
        # not owner cannot rebuild
        yield self.assertUserForbidden("builds/13", "rebuild", {}, "eightuser")

        # can rebuild build with matching builder
        allow_rules = [
            RebuildBuildEndpointMatcher(role="eight-*", builder="mybuilder"),
            AnyEndpointMatcher(role="admins"),
        ]
        self.setAllowRules(allow_rules)
        yield self.assertUserAllowed("builds/13", "rebuild", {}, "eightuser")
        yield self.assertUserForbidden("builds/999", "rebuild", {}, "eightuser")

        # cannot rebuild build with non-matching builder
        allow_rules = [
            RebuildBuildEndpointMatcher(role="eight-*", builder="foo"),
            AnyEndpointMatcher(role="admins"),
        ]
        self.setAllowRules(allow_rules)
        yield self.assertUserForbidden("builds/13", "rebuild", {}, "eightuser")

    @defer.inlineCallbacks
    def test_fnmatchPatternRoleCheck(self) -> InlineCallbacksType[None]:
        # defaultDeny is True by default so action is denied if no match
        allow_rules: list[EndpointMatcherBase] = [AnyEndpointMatcher(role="[a,b]dmin?")]

        self.setAllowRules(allow_rules)

        yield self.assertUserAllowed("builds/13", "rebuild", {}, "homer")

        # check if action is denied
        with self.assertRaisesRegex(authz.Forbidden, '403 you need to have role .+'):
            yield self.assertUserAllowed("builds/13", "rebuild", {}, "nineuser")

        with self.assertRaisesRegex(authz.Forbidden, '403 you need to have role .+'):
            yield self.assertUserAllowed("builds/13", "rebuild", {}, "eightuser")

    @defer.inlineCallbacks
    def test_regexPatternRoleCheck(self) -> InlineCallbacksType[None]:
        # change matcher
        self.authz.match = authz.reStrMatcher

        # defaultDeny is True by default so action is denied if no match
        allow_rules: list[EndpointMatcherBase] = [
            AnyEndpointMatcher(role="(admin|agent)s"),
        ]

        self.setAllowRules(allow_rules)

        yield self.assertUserAllowed("builds/13", "rebuild", {}, "homer")
        yield self.assertUserAllowed("builds/13", "rebuild", {}, "bond")

        # check if action is denied
        with self.assertRaisesRegex(authz.Forbidden, '403 you need to have role .+'):
            yield self.assertUserAllowed("builds/13", "rebuild", {}, "nineuser")

        with self.assertRaisesRegex(authz.Forbidden, '403 you need to have role .+'):
            yield self.assertUserAllowed("builds/13", "rebuild", {}, "eightuser")

    @defer.inlineCallbacks
    def test_DefaultDenyFalseContinuesCheck(self) -> InlineCallbacksType[None]:
        # defaultDeny is True in the last rule so action is denied in the last check
        allow_rules: list[EndpointMatcherBase] = [
            AnyEndpointMatcher(role="not-exists1", defaultDeny=False),
            AnyEndpointMatcher(role="not-exists2", defaultDeny=False),
            AnyEndpointMatcher(role="not-exists3", defaultDeny=True),
        ]

        self.setAllowRules(allow_rules)
        # check if action is denied and last check was exact against not-exist3
        with self.assertRaisesRegex(authz.Forbidden, '.+not-exists3.+'):
            yield self.assertUserAllowed("builds/13", "rebuild", {}, "nineuser")

    @defer.inlineCallbacks
    def test_DefaultDenyTrueStopsCheckIfFailed(self) -> InlineCallbacksType[None]:
        # defaultDeny is True in the first rule so action is denied in the first check
        allow_rules: list[EndpointMatcherBase] = [
            AnyEndpointMatcher(role="not-exists1", defaultDeny=True),
            AnyEndpointMatcher(role="not-exists2", defaultDeny=False),
            AnyEndpointMatcher(role="not-exists3", defaultDeny=False),
        ]

        self.setAllowRules(allow_rules)

        # check if action is denied and last check was exact against not-exist1
        with self.assertRaisesRegex(authz.Forbidden, '.+not-exists1.+'):
            yield self.assertUserAllowed("builds/13", "rebuild", {}, "nineuser")
